#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.
import pandas as pd
import pytest

from ..example import MitoWidget, sheet
from ..utils import get_invalid_headers
from mitosheet.errors import EditError

def test_example_creation_blank():
    df = pd.DataFrame()
    w = MitoWidget(df)

VALID_DATAFRAMES = [
    (pd.DataFrame()),
    (pd.DataFrame(data={'A': [1, 2, 3]})),
    (pd.DataFrame(data={'A0123': [1, 2, 3]})),
]
@pytest.mark.parametrize("df", VALID_DATAFRAMES)
def test_sheet_creates_valid_dataframe(df):
    mito = sheet(df)
    assert mito is not None

NON_VALID_HEADER_DATAFRAMES = [
    (pd.DataFrame(data={0: [1, 2, 3]}), ['0_']),
    (pd.DataFrame(data={0.1: [1, 2, 3]}), ['0_1_']),
    (pd.DataFrame(data={'A': [1, 2, 3], 0: [1, 2, 3]}), ['A', '0_']),
    (pd.DataFrame(data={'A': [1, 2, 3], '000': [1, 2, 3]}), ['A', '000_']),
    (pd.DataFrame(data={'A': [1, 2, 3], 'abc-123': [1, 2, 3]}), ['A', 'abc_123']),
    (pd.DataFrame(data={'A': [1, 2, 3], '-123': [1, 2, 3]}), ['A', '_123']),
    (pd.DataFrame(data={'A': [1, 2, 3], '-123_': [1, 2, 3]}), ['A', '_123_']),
    (pd.DataFrame(data={'A': [1, 2, 3], '-abc_': [1, 2, 3]}), ['A', '_abc_']),
    (pd.DataFrame(data={'123': [1, 2, 3], '-abc_': [1, 2, 3]}), ['123_', '_abc_']),
    (pd.DataFrame(data={'ABC!DEF': [1, 2, 3], '123': [1, 2, 3]}), ['ABC_DEF', '123_']),
    (pd.DataFrame(data={'ABC?DEF': [1, 2, 3], '123': [1, 2, 3]}), ['ABC_DEF', '123_']),
    (pd.DataFrame(data={' ': [1, 2, 3], '123': [1, 2, 3]}), ['_', '123_']),
    (pd.DataFrame(data={'-': [1, 2, 3], ' !': [1, 2, 3]}), ['_', '__']),
    (pd.DataFrame(data={'##': [1, 2, 3], ' !': [1, 2, 3]}), ['numnum', '__']),
    (pd.DataFrame(data={'#!?5': [1, 2, 3], '#123': [1, 2, 3]}), ['num__5', 'num123']),
    (pd.DataFrame(data={'()': [1, 2, 3], '.,,': [1, 2, 3]}), ['__', '___']),
    (pd.DataFrame(data={'nate': [1, 2, 3], '.,,': [1, 2, 3]}), ['nate', '___']),
    (pd.DataFrame(data={'//': [1, 2, 3], '-': [1, 2, 3]}), ['__', '_']),
]

@pytest.mark.parametrize("df, new_keys", NON_VALID_HEADER_DATAFRAMES)
def test_sheet_errors_during_non_string_headers(df, new_keys):
    assert len(get_invalid_headers(df)) != 0
    mito = sheet(df)
    assert list(mito.widget_state_container.curr_step['dfs'][0].keys()) == new_keys

def test_create_with_multiple_dataframes():
    mito = sheet(pd.DataFrame(), pd.DataFrame(), pd.DataFrame(), pd.DataFrame(data={'A': [1, 2, 3]}))
    assert mito is not None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A reorder_column step, which allows you to move 
a column to a different location in the df.
"""
from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
)

REORDER_COLUMN_EVENT = 'reorder_column_edit'
REORDER_COLUMN_STEP_TYPE = 'reorder_column'

REORDER_COLUMN_PARAMS = [
    'sheet_index', # int
    'column_header', # the column to reorder
    'new_column_index', # the new location for the column
]

def execute_reorder_column_step(
        wsc,
        sheet_index,
        column_header,
        new_column_index
    ):
    """
    The function responsible for updating the widget state container
    with a new column reorder step.

    If it fails part of the way through, deletes the new column reorder step entirely.
    """

    # if the sheet doesn't exist, throw an error
    if not wsc.does_sheet_index_exist(sheet_index):
        raise make_no_sheet_error(sheet_index)

    # We check that the column to be reordered exists
    missing_column = set([column_header]).difference(wsc.curr_step['column_metatype'][sheet_index].keys())
    if len(missing_column) > 0:
        raise make_no_column_error(missing_column)

    # make sure new_column_index is valid
    if new_column_index < 0:
        new_column_index = 0

    if new_column_index >= len(wsc.curr_step['dfs'][sheet_index].columns):
        new_column_index = len(wsc.curr_step['dfs'][sheet_index].columns) - 1
        
    # Create a new step
    wsc._create_and_checkout_new_step(REORDER_COLUMN_STEP_TYPE)
    # Save the params of the current step
    wsc.curr_step['sheet_index'] = sheet_index
    wsc.curr_step['column_header'] = column_header
    wsc.curr_step['new_column_index'] = new_column_index

    # Actually execute the column reordering
    wsc.curr_step['dfs'][sheet_index] = _execute_reorder_column(
        wsc.curr_step['dfs'][sheet_index],
        column_header,
        new_column_index
    )

def _execute_reorder_column(df, column_header, new_column_index):
    """
    Helper function for reordering a column in the dataframe
    """
    df_columns = [col for col in df.columns if col != column_header]
    df_columns.insert(new_column_index, column_header)
    return df[df_columns]


def transpile_reorder_column_step(
        widget_state_container,
        step,
        sheet_index,
        column_header,
        new_column_index
    ):
    """
    Transpiles a column reorder step to python code!
    """

    # get columns in df
    columns_list_line = f'{step["df_names"][sheet_index]}_columns = [col for col in {step["df_names"][sheet_index]}.columns if col != \'{column_header}\']'

    # insert column into correct location 
    insert_line = f'{step["df_names"][sheet_index]}_columns.insert({new_column_index}, \'{column_header}\')'
    
    # Apply reorder line
    apply_reorder_line = f'{step["df_names"][sheet_index]} = {step["df_names"][sheet_index]}[{step["df_names"][sheet_index]}_columns]'

    return [columns_list_line, insert_line, apply_reorder_line]


REORDER_COLUMN_STEP = {
    'step_version': 1,
    'event_type': REORDER_COLUMN_EVENT,
    'step_type': REORDER_COLUMN_STEP_TYPE,
    'params': REORDER_COLUMN_PARAMS,
    'execute': execute_reorder_column_step,
    'transpile': transpile_reorder_column_step
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.
import pandas as pd
import pytest
import json

from mitosheet.tests.test_utils import create_mito_wrapper_dfs


def test_sheet_json_holds_dates_correctly():
    df = pd.DataFrame({
        'name': ['alice','bob','charlie'],
        'date_of_birth': ['2005-10-25','2002-10-2','2001-11-14']
    })

    df['date_of_birth'] = pd.to_datetime(df['date_of_birth'])

    mito = create_mito_wrapper_dfs(df)
    
    sheet_json = json.loads(mito.mito_widget.sheet_json)
    assert sheet_json[0]['data'][0][1] == '2005-10-25 00:00:00'
    assert sheet_json[0]['data'][1][1] == '2002-10-02 00:00:00'
    assert sheet_json[0]['data'][2][1] == '2001-11-14 00:00:00'



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A raw python import, which allows you to import dataframes with the 
given Python code
"""
from mitosheet.utils import get_header_renames
from mitosheet.errors import make_df_exists_error
import json

RAW_PYTHON_IMPORT_EVENT = 'raw_python_import_edit'
RAW_PYTHON_IMPORT_STEP_TYPE = 'raw_python_import'

RAW_PYTHON_IMPORT_PARAMS = [
    'python_code', # a big string
    'new_df_names', # df names
]

def execute_raw_python_import(
        wsc,
        python_code,
        new_df_names
    ):
    """
    Imports the python code - and makes sure to add the corresponding dataframes
    to the sheet itself
    """

    for df_name in new_df_names:
        if df_name in wsc.curr_step['df_names']:
            raise make_df_exists_error(df_name)

    # Create a new step and save the inputs
    wsc._create_and_checkout_new_step(RAW_PYTHON_IMPORT_STEP_TYPE)
    wsc.curr_step['python_code'] = python_code
    wsc.curr_step['new_df_names'] = new_df_names

    # Build a global string so these variables are defined in the outer scope (e.g. this executing code)
    global_string = ';'.join([f'global {df_name}' for df_name in new_df_names])
    # Actually run the code to define the variables
    exec(f'{global_string};{python_code}')
    # And then get the dataframes out of global storage (renaming any headers we need to along the way)
    column_header_renames = dict()
    for df_name in new_df_names:
        df = globals()[df_name]
        # If there are any invalid names, we perform a rename
        renames = get_header_renames(df.keys())
        if len(renames) > 0:
            # Save that we did these renames
            column_header_renames[df_name] = renames
            # Actually perform any renames we need to
            df.rename(columns=renames, inplace=True)

        wsc.add_df_to_curr_step(df, df_name=df_name)
    
    # Save the renames that have occured in the step, for transpilation reasons
    wsc.curr_step['column_header_renames'] = column_header_renames


def transpile_raw_python_import(
        widget_state_container,
        step,
        python_code,
        new_df_names
    ):
    """
    Transpiles a raw Python import to Python code. Because we're giving raw Python code,
    this is pretty easy - just split the code at the line breaks - and then
    perform any renames we need to.
    """
    code = python_code.split('\n')

    # Perform any renames we need to
    for df_name in step['column_header_renames']:
        renames = step['column_header_renames'][df_name]
        code.append(
            f'{df_name}.rename(columns={json.dumps(renames)}, inplace=True)'
        )

    return code


RAW_PYTHON_IMPORT_STEP = {
    'step_version': 1,
    'event_type': RAW_PYTHON_IMPORT_EVENT,
    'step_type': RAW_PYTHON_IMPORT_STEP_TYPE,
    'params': RAW_PYTHON_IMPORT_PARAMS,
    'execute': execute_raw_python_import,
    'transpile': transpile_raw_python_import
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains handlers for the Mito API
"""

from mitosheet.steps.import_steps.raw_python_import import RAW_PYTHON_IMPORT_STEP_TYPE
from mitosheet.steps.import_steps.simple_import import SIMPLE_IMPORT_STEP_TYPE
from mitosheet.save_utils import read_analysis
from mitosheet.mito_analytics import analytics, log_dfs_metadata, static_user_id
import os
from pathlib import Path

def get_filenames_with_suffix(suffix):
    """
    Returns all the file names in the current folder that end with the given
    suffix, sorted from most-recently created to the oldest.
    """
    # We sort them by creation time, to get the most recent files, as the user
    # is more likely to want these
    filenames = sorted(Path('.').iterdir(), key=os.path.getmtime)
    filenames.reverse()
    return [str(filename) for filename in filenames if filename.suffix == suffix]


def handle_datafiles(send, event):
    """
    Handles a `datafiles` api call, and returns all the csv files
    in the current folder.
    """
    csv_files = get_filenames_with_suffix('.csv')
    # TODO: also get the XLSX files, when we can import them
    send({
        'event': 'api_response',
        'id': event['id'],
        'data': csv_files
    })

def handle_import_summary(send, event):
    """
    Handle import summary is a route that, given the name of an analysis, will
    return the parameters to import steps over the course of the analysis. 

    The data we return is in the format:
    {
        "1": {
            "file_names": ["file123.csv"]
        }, 
        "3": {
            "python_code": "import pandas as ...",
            "new_df_names": ["df1"]
        }
    }
    which is a mapping from raw import steps to the files that they import.
    """
    analysis_name = event['analysis_name']
    analysis = read_analysis(analysis_name)

    imports_only = dict()
    if analysis is not None:
        for step_id, step in analysis['steps'].items():
            if step['step_type'] == SIMPLE_IMPORT_STEP_TYPE:
                imports_only[step_id] = dict()
                imports_only[step_id]['step_type'] = SIMPLE_IMPORT_STEP_TYPE
                imports_only[step_id]['file_names'] = step['file_names']
            elif step['step_type'] == RAW_PYTHON_IMPORT_STEP_TYPE:
                imports_only[step_id] = dict()
                imports_only[step_id]['step_type'] = RAW_PYTHON_IMPORT_STEP_TYPE
                imports_only[step_id]['python_code'] = step['python_code']
                imports_only[step_id]['new_df_names'] = step['new_df_names']

    send({
        'event': 'api_response',
        'id': event['id'],
        'data': imports_only
    })

def handle_api_event(send, event):
    """
    Handler for all API calls. Note that any response to the
    API must return the same ID that the incoming message contains,
    so that the frontend knows how to match the responses.
    """
    # Log the api call
    analytics.track(static_user_id, event['type'], event)
    
    # And then handle it
    if event['type'] == 'datafiles':
        handle_datafiles(send, event)
    elif event['type'] == 'import_summary':
        handle_import_summary(send, event)
    else:
        raise Exception(f'Event: {event} is not a valid API call')


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains helpful utility functions
"""
import json
import pandas as pd
import numpy as np
import numbers
from string import ascii_letters, digits

ALLOWED_CHARS = ascii_letters + digits + '_'


def make_valid_header(column_header):
    """
    Takes a header, and performs replaces against common characters
    to make the column_header valid!
    """
    # If it's just numbers, turn it into a string (with an underscore)
    if isinstance(column_header, numbers.Number):
        return str(column_header).replace('.', '_') + '_'

    # If it's just numbers in a string, add an underscore
    if column_header.isdigit():
        return column_header + "_"

    # Replace common invalid seperators with valid seperators
    replace_dict = {
        ' ': '_',
        '-': '_',
        '(': '_',
        ')': '_',
        '/': '_',
        '#': 'num',
        ',': '_',
        '.': '_',
        '!': '_',
        '?': '_'
    }
    for find, replace in replace_dict.items():
        column_header = column_header.replace(find, replace)
    
    if not is_valid_header(column_header):
        # If our reasonable replacements don't work, we just ban any character not from a set
        new_header = ''.join([
            c for c in column_header if c in ALLOWED_CHARS
        ])
        # And then append an underscore, for good measure
        new_header = new_header + '_'

        return new_header
    return column_header

def is_valid_header(column_header):
    """
    A header is valid if:
    1. It is a string
    2. It only contains alphanumber characters, or _
    3. It has at least one non-numeric character.

    Valid examples: A, ABC, 012B, 213_bac, 123_123
    Invalid examples: 123, 123!!!, ABC!, 123-123

    This is a result of not being able to distingush column headers from constants
    otherwise, and would not be necessary if we had a column identifier!
    """

    return isinstance(column_header, str) and ( # Condition (1)
        set(column_header).issubset(set(ascii_letters).union(set(digits)).union(set(['_']))) and # Condition (2)
        not column_header.isdigit() # Condition (3)
    )


def get_invalid_headers(df: pd.DataFrame):
    """
    Given a dataframe, returns a list of all the invalid headers this list has. 
    """
    return [
        header for header in df.columns.tolist()
        if not is_valid_header(header)
    ]

def get_header_renames(column_headers):
    """
    Given a list of column headers, returns a mapping from old, invalid headers to
    new, valid headers. Empty if no renames are necessary.
    """
    renames = dict()
    for column_header in column_headers:
        if not is_valid_header(column_header):
            valid_header = make_valid_header(column_header)
            renames[column_header] = valid_header

    return renames


def dfs_to_json(dfs):
    return json.dumps([df_to_json_dumpsable(df) for df in dfs])


def df_to_json_dumpsable(df: pd.DataFrame):
    """
    Returns a dataframe represented in a way that can be turned into a 
    JSON object with json.dumps
    """
    # We make a copy, as we modify the df below
    df = df.copy(deep=True).head(n=2000) # we only show the first 2k rows!
    
    # First, we figure out which of the columns contain dates, and we
    # convert them to string columns (for formatting reasons).
    # NOTE: we don't use date_format='iso' in df.to_json call as it appends seconds to the object, 
    # see here: https://stackoverflow.com/questions/52730953/pandas-to-json-output-date-format-in-specific-form
    date_columns = df.select_dtypes(include=['datetime64'])
    for key in date_columns:
        df[key] = date_columns[key].dt.strftime('%Y-%m-%d %X')

    json_obj = json.loads(df.to_json(orient="split"))
    # Then, we go through and find all the null values (which are infinities),
    # and set them to 'NaN' for display in the frontend.
    for d in json_obj['data']:
        for idx, e in enumerate(d):
            if e is None:
                d[idx] = 'NaN'

    return json_obj


def get_column_filter_type(series):
    """
    Given a series, take's the series dtype and catergorizes the correct
    column_filter_type so that the filter modal can display only
    valid filtering options
    
    TODO: unify this with the other Mito types!
    """
    try: 
        if series.dtype == np.float64 or series.dtype == np.int64:
            return 'number'
        elif series.dtype == 'datetime64[ns]':
            return 'datetime'
        elif series.dtype == np.object:
            return 'string'
    except:
        # If there is an error, we just call this a string column. In the wild, we have
        # seen errors where dtype is not defined on the df object, but we have been
        # unable to replicate this or figure out when it occurs.
        return 'string'
    
    # TODO: what do we do if it's none? for now just return string, for no reason.
    return 'string'
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains tests for edit events.
"""
from typing import List

from mitosheet.steps import ADD_COLUMN_STEP, STEPS
from mitosheet.steps import DELETE_COLUMN_STEP
from mitosheet.steps import RENAME_COLUMN_STEP
from mitosheet.steps import FILTER_STEP
from mitosheet.steps import GROUP_STEP
from mitosheet.steps import MERGE_STEP
from mitosheet.steps import REORDER_COLUMN_STEP
from mitosheet.steps import SET_COLUMN_FORMULA_STEP
from mitosheet.steps import SORT_STEP
from mitosheet.steps.import_steps.simple_import import SIMPLE_IMPORT_STEP
from mitosheet.steps.import_steps.raw_python_import import RAW_PYTHON_IMPORT_STEP


def check_step(
        step_definition, 
        step_version: int, 
        step_type: str, 
        params: List[str]
    ):
    """
    Helper function that checks a given step definition against the 
    expected step_version, step_type, and params for that step. 

    Throws an assertion error if any of them do not match! 
    """
    assert step_definition['step_version'] == step_version
    assert step_definition['step_type'] == step_type
    assert step_definition['params'] == params

def test_params_static():
    """
    NOTE: This is a regression test! Before changing this test to make it pass, talk to 
    the engineering team and make sure you know what you're doing. 

    Remeber:
    1. Each Mito analysis is written to a file, where each step is saved with its
       parameters.
    2. If the _name_ of the step, or the _parameters to the step_ change, then this
       will break existing user analyses. 
    
    Thus, this test is to make sure that we _know_ when we're breaking things. 
    
    However, note that there are ways to break existing user analyses other than this. 
    For example, if you change how the group step flattens headers, this may cause
    issues if the user then later renames one of those flattened columns. So, this
    regression test is not perfect, but it is a good start!
    """

    check_step(
        ADD_COLUMN_STEP,
        1,
        'add_column',
        ['sheet_index', 'column_header']
    )

    check_step(
        DELETE_COLUMN_STEP,
        1,
        'delete_column',
        ['sheet_index', 'column_header']
    )

    check_step(
        RENAME_COLUMN_STEP,
        1,
        'rename_column',
        ['sheet_index', 'old_column_header', 'new_column_header']
    )

    check_step(
        FILTER_STEP,
        1,
        'filter_column',
        ['sheet_index', 'column_header', 'operator', 'filters']
        # TODO: this param "filters" is really weakly typed here, and could totally
        # change without being detected by this test. Boo!
    )
    
    check_step(
        GROUP_STEP,
        1,
        'group',
        ['sheet_index', 'group_rows', 'group_columns', 'values']
        # TODO: the param "values" is also very weakly typed, and could change
        # without this test detecting it
    )

    check_step(
        MERGE_STEP,
        1,
        'merge',
        ['sheet_index_one', 'merge_key_one', 'selected_columns_one', 'sheet_index_two', 'merge_key_two', 'selected_columns_two']
    )

    check_step(
        REORDER_COLUMN_STEP,
        1,
        'reorder_column',
        ['sheet_index', 'column_header', 'new_column_index']
    )

    check_step(
        SET_COLUMN_FORMULA_STEP,
        1,
        'set_column_formula',
        ['sheet_index', 'column_header', 'old_formula', 'new_formula']
    )

    check_step(
        SORT_STEP,
        1,
        'sort',
        ['sheet_index', 'column_header', 'sort_direction']
    )

    check_step(
        SIMPLE_IMPORT_STEP,
        1,
        'simple_import',
        ['file_names']
    )

    check_step(
        RAW_PYTHON_IMPORT_STEP,
        1,
        'raw_python_import',
        ['python_code', 'new_df_names']
    )

    assert len(STEPS) == 11






#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A sort step, which allows you to sort a df based on some key column
"""
import pandas as pd

from mitosheet.errors import (
    EditError,
    make_execution_error,
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_sort_error
)

SORT_EVENT = 'sort_edit'
SORT_STEP_TYPE = 'sort'

SORT_COLUMN_PARAMS = [
    'sheet_index', # int
    'column_header', # column to sort
    'sort_direction' # string either 'ascending' or 'descending'
]

# CONSTANTS USED IN THE SORT STEP ITSELF
ASCENDING = 'ascending'
DESCENDING = 'descending'

def execute_sort(
        wsc,
        sheet_index,
        column_header,
        sort_direction,
    ):
    """
    sorts an existing sheet with the given sort_direction on the column_header
    """
    # if the sheet doesn't exist, throw an error
    if not wsc.does_sheet_index_exist(sheet_index):
        raise make_no_sheet_error(sheet_index)

    # We check that the sorted column exists 
    missing_column = set([column_header]).difference(wsc.curr_step['column_metatype'][sheet_index].keys())
    if len(missing_column) > 0: 
        raise make_no_column_error(missing_column)

    # If no errors we create a new step for this sort
    wsc._create_and_checkout_new_step(SORT_STEP_TYPE)
    # update the step to save the variables needed to reconstruct the sort
    wsc.curr_step['sheet_index'] = sheet_index
    wsc.curr_step['column_header'] = column_header
    wsc.curr_step['sort_direction'] = sort_direction

    # execute on real dataframes
    wsc.dfs[sheet_index] = _execute_sort(
        wsc.dfs[sheet_index], 
        column_header,
        sort_direction
    )

def _execute_sort(
        df,
        column_header,
        sort_direction,
    ):
    """
    Executes a sort on the given df, by sorting the column named 
    column_header in sort_direction (ascending or descending)
    """

    try: 
        new_df = df.sort_values(by=column_header, ascending=(sort_direction == ASCENDING), na_position='first')
        return new_df.reset_index(drop=True)
    except TypeError as e:
        # A NameError occurs when you try to sort a column with incomparable 
        # dtypes (ie: a column with strings and floats)
        print(e)
        # Generate an error informing the user
        raise make_invalid_sort_error(column_header)


def transpile_sort(
        widget_state_container,
        step,
        sheet_index,
        column_header,
        sort_direction
    ):
    """
    Transpiles a sort step to Python code. 
    """
    # sort the dataframe
    df_name = step["df_names"][sheet_index]
    sort_line = f'{df_name} = {df_name}.sort_values(by=\'{column_header}\', ascending={sort_direction == ASCENDING}, na_position=\'first\')'
    reset_index_line = f'{df_name} = {df_name}.reset_index(drop=True)'

    return [sort_line, reset_index_line]


SORT_STEP = {
    'step_version': 1,
    'event_type': SORT_EVENT,
    'step_type': SORT_STEP_TYPE,
    'params': SORT_COLUMN_PARAMS,
    'execute': execute_sort,
    'transpile': transpile_sort
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains @decorators to help Mito sheet functions return and handle
types in a consistent way.
"""
import pandas as pd
import numpy as np
from functools import wraps
from mitosheet.errors import make_invalid_arguments_error, make_function_execution_error

# Because type(1) = int, thus 1 is a 'number' in the Mito type system
MITO_PRIMITIVE_TYPE_MAPPING = {
    'number': [int, float],
    'string': [str],
    'boolean': [bool],
    'timestamp': [pd.Timestamp]
}

# NOTE: these are the dtype strings of series used by pandas!
MITO_SERIES_TYPE_MAPPING = {
    'number_series': ['int64', 'float64'],
    'string_series': ['str', 'object'],
    'boolean_series': ['bool'],
    'datetime_series': ['datetime64[ns]']
}

MITO_TYPE_PRIMITIVE_TO_SERIES_MAPPING = {
    'number': 'number_series',
    'string': 'string_series',
    'boolean': 'boolean_series',
    'timestamp': 'datetime_series'
}

def identity(x):
    """
    Helper function for the MITO_TYPE_CONVERSION_MAP, when turning from the conversion is from
    one type to itself (and so no work needs to be done).
    """
    return x


"""
A mapping of functions from converting from one Mito type to another. 

MITO_TYPE_CONVERSION_MAP['number']['string'] returns a function that 
converts from a number to a string. 

If there is no conversion possible the above is defined as None!

NOTE: the conversion functions may error and this should be handeled!
"""
MITO_TYPE_CONVERSION_MAP = {
    'number': {
        'number': identity,
        'string': lambda num: str(num),
        'boolean': lambda num: bool(num),
        'timestamp': None,
        'number_series': lambda num: pd.Series(data=[num]),
        'string_series': lambda num: pd.Series(data=[str(num)]),
        'boolean_series': lambda num: pd.Series(data=[bool(num)]),
        'datetime_series': None
    },
    'string': {
        'number': lambda string: float(string),
        'string': identity,
        'boolean': lambda string: bool(string),
        'timestamp': None,
        'number_series': lambda string: pd.Series(data=[float(string)]),
        'string_series': lambda string: pd.Series(data=[string]),
        'boolean_series': lambda string: pd.Series(data=[bool(string)]),
        'datetime_series': lambda string: pd.Series(data=[pd.to_datetime(string)], dtype='datetime64[ns]')
    },
    'boolean': {
        'number': lambda boolean: float(boolean),
        'string': lambda boolean: str(boolean),
        'boolean': identity,
        'timestamp': None,
        'number_series': lambda boolean: pd.Series(data=[float(boolean)]),
        'string_series': lambda boolean: pd.Series(data=[str(boolean)]),
        'boolean_series': lambda boolean: pd.Series(data=[boolean]),
        'datetime_series': None
    },
    'timestamp': {
        'number': None,
        'string': None,
        'boolean': None,
        'timestamp': identity,
        'number_series': None,
        'string_series': lambda timestamp: pd.Series(data=[timestamp.strftime('%Y-%m-%d %X')]),
        'boolean_series': None,
        'datetime_series': lambda timestamp: pd.Series(data=[timestamp], dtype='datetime64[ns]')
    },
    'number_series': {
        'number': None,
        'string': None,
        'boolean': None,
        'timestamp': None,
        'number_series': identity,
        'string_series': lambda num_series: num_series.astype('str'),
        'boolean_series': lambda num_series: num_series.astype('bool'),
        'datetime_series': None
    },
    'string_series': {
        'number': None,
        'string': None,
        'boolean': None,
        'timestamp': None,
        'number_series': lambda str_series: str_series.astype('float64'),
        'string_series': lambda str_series: str_series.astype('str'),
        'boolean_series': lambda str_series: str_series.astype('bool'),
        'datetime_series': lambda str_series: pd.to_datetime(str_series)
    },
    'boolean_series': {
        'number': None,
        'string': None,
        'boolean': None,
        'timestamp': None,
        'number_series': lambda bool_series: bool_series.astype('float64'),
        'string_series': lambda bool_series: bool_series.astype('str'),
        'boolean_series': identity,
        'datetime_series': None
    },
    'datetime_series': {
        'number': None,
        'string': None,
        'boolean': None,
        'timestamp': None,
        'number_series': None,
        'string_series': lambda datetime_series: datetime_series.dt.strftime('%Y-%m-%d %X'),
        'boolean_series': None,
        'datetime_series': identity
    }
}

# Supertypes are the parent types of other types. There is only one currently:
# 'series' has subtypes that end in _series.
SUPERTYPES = set(['series'])

MITO_TYPES = set(MITO_PRIMITIVE_TYPE_MAPPING.keys()).union(MITO_SERIES_TYPE_MAPPING.keys()).union(SUPERTYPES)


def get_mito_type(obj):

    if isinstance(obj, pd.Series):
        dtype = obj.dtype
        for key, value in MITO_SERIES_TYPE_MAPPING.items():
            if dtype in value:
                return key

    elif isinstance(obj, pd.Timestamp):
        return 'timestamp'
    else:
        obj_type = type(obj)

        for key, value in MITO_PRIMITIVE_TYPE_MAPPING.items():
            if obj_type in value:
                return key

    return None

def is_in_type_set(type_set, obj):
    mito_type = get_mito_type(obj)

    if mito_type is None:
        return False

    return mito_type in type_set


def as_types(
        *args, 
        remaining=False, 
        ignore_conversion_failures=False, 
        num_optional=0,
        attempt_output_types=None
    ):
    """
    This generator factory will attempt to take the given arguments to the decorated sheet
    function and cast them to type arguments given in the decorator.

    This will primarily error if:
    0. If the wrong number of types are passed to the sheet function.
    1. A conversion cannot be made from the passed type to the target type:
        - You can't convert from a string_series to a string, for example
        - You can't convert from the string "hi" to a number
        - See the mappings defined above in the MITO_TYPE_CONVERSION_MAP above!

    However, if ignore_conversion_failures is True, will ignore errors of type (1), 
    and instead will just leave out the arguments that fail to convert.

    attempt_output_types is an optional parameter. If passed, it should be a function
    that takes all arguments as input, and returns the Mito type that it wants the result
    to be returned as.
    """

    types = args

    if set(args).difference(MITO_TYPES):
        raise Exception(f'Types to as_types must be from {MITO_TYPES}, not {args}')

    def wrap(sheet_function):
        @wraps(sheet_function)
        def wrapped_f(*args):
            # First, we check that the number of arguments passed is correct
            if not remaining:
                # If not remaining, then the function has a specific number of arguments
                if len(types) > len(args) + num_optional:
                    raise make_invalid_arguments_error(sheet_function.__name__)

            else:
                # Otherwise, the function takes a variable number of arguments, so we must have less
                # arguments than the number of given types
                if len(types) > len(args):
                    raise make_invalid_arguments_error(sheet_function.__name__)

            # before changing the types, determine the indexes that contain nan values
            nan_indexes = get_nan_indexes(*args)

            # if the sheet_function is FILLNAN, then don't remove the NaN values. 
            if str(sheet_function.__name__) == 'FILLNAN':
                nan_indexes = [False for i in range(len(args[0]))]
                
            # for each arg that is an instance of a pd.series, filter out the nan_indexes
            new_args = []
            for arg in args:
                if isinstance(arg, pd.Series): 
                    # only keep the rows at indexes that don't contain a nan_index in any of the
                    # series arguments to the sheet function
                    processed_arg = arg.loc[[not boolean for boolean in nan_indexes]]

                    # then reset the indexes so that when the sheet function creates a new series
                    # it will have the same indexes as the passed in series

                    # drop = True Just resets the index, without inserting it as a column in the new DataFrame.
                    processed_arg = processed_arg.reset_index(drop=True)
                    new_args.append(processed_arg)

                else:
                    new_args.append(arg)

            # Then, we fill the types to make sure there is a type for each of the passed arguments
            filled_types = [types[index] if index < len(types) else types[-1] for index, _ in enumerate(new_args)]
            new_arguments = []

            # Finally, we check that each given argument is of the valid type
            for to_type, argument in zip(filled_types, new_args):
                from_type = get_mito_type(argument)

                # If we just want it as a series, we convert it whatever type of
                # series doesn't change the type of data
                if to_type == 'series':
                    if from_type.endswith('series'):
                        # If it's already a series, we don't need to convert
                        new_arguments.append(argument)
                        continue
                    to_type = MITO_TYPE_PRIMITIVE_TO_SERIES_MAPPING[from_type]

                # We actually try and make a conversion
                conversion_function = MITO_TYPE_CONVERSION_MAP[from_type][to_type]

                # Error if no conversion is possible
                if conversion_function is None:
                    if not ignore_conversion_failures:
                        # Occurs if you cannot convert from {from_type} to {to_type}
                        raise make_invalid_arguments_error(sheet_function.__name__)

                    else:
                        continue

                try:
                    converted_arg = conversion_function(argument)

                except Exception as e:
                    if not ignore_conversion_failures:
                        # Occurs if you cannot convert from {from_type} to {to_type}

                        raise make_invalid_arguments_error(sheet_function.__name__)
                    else:
                        continue
                
                # If we can convert, then save the conversion as an argument!
                new_arguments.append(converted_arg)
                
            try:
                # actually apply the sheet function
                result = sheet_function(*new_arguments)

                # If we should cast the output, we try to cast it
                if attempt_output_types is not None:
                    attempt_output_type = attempt_output_types(*new_args)
                    # If there is no conversion available, we give up
                    if attempt_output_type is not None:

                        original_output_type = get_mito_type(result)
                        final_conversion_function = MITO_TYPE_CONVERSION_MAP[original_output_type][attempt_output_type]

                        if final_conversion_function is not None:
                            # NOTE: This can error during the conversion, in which case an error
                            # is reported back to the user that this function failed...
                            result = final_conversion_function(result)

                            # add back the nan values
                            return put_nan_indexes_back(result, nan_indexes) 

                # If we don't want to cast it, or if no cast is available, we just add back the nan values and return 
                return put_nan_indexes_back(result, nan_indexes)
            except:
                raise make_function_execution_error(sheet_function.__name__)

            return result
        return wrapped_f
    return wrap


def get_series_type_of_first_arg(*args):
    """
    Given a tuple of arguments, returns the Mito type of the first argument,
    in it's series interpretation.

    Useful for being passed as attempt_output_types for functions like 
    LEFT, RIGHT, MID, etc - where we want the function to preserve
    the output types if possible.
    """ 
    if len(args) == 0:
        return None

    mito_type = get_mito_type(args[0])

    # If it's a primitive type, we want it to be returned as a series
    if mito_type in MITO_PRIMITIVE_TYPE_MAPPING.keys():
        return MITO_TYPE_PRIMITIVE_TO_SERIES_MAPPING[mito_type]
    else:
        return mito_type


def get_nan_indexes(*argv): 
    """
    Given the list of arguments to a function, as *argv, 
    returns a list of row indexes that is True iff one of the series 
    params is NaN at that index. Otherwise the index is False

    This function is called by the as_types decorator 
    """

    # we find the max length of the args because we are unsure ahead of time which arg 
    # is a series. we need the max_length to construct the boolean index array
    max_length = -1
    for arg in argv:
        # check the type to make sure an error is not thrown for calling len() on an int
        if isinstance(arg, pd.Series) and len(arg) > max_length:
            max_length = len(arg)

    # if there are no series, then:
    #   1. there are no NaN values
    #   2. all of the args are a length of 1
    if max_length == -1:
        return [False]

    nan_indexes = [False for i in range(max_length)]

    # for each row, check for NaN values in the function
    for arg in argv:
        if isinstance(arg, pd.Series): 
            for idx in range(len(arg)):
                if pd.isna(arg[idx]):
                    nan_indexes[idx] = True

    return nan_indexes


def put_nan_indexes_back(series, nan_indexes):
    """
    This functions takes a series and a boolean index list that is True 
    if the index in the series should be NaN and false if it should
    be left alone. 

    Returns the series with the NaN values put in

    This function is called by the as_types decorator 
    """
    original_length = len(nan_indexes)
    final_series = []
    real_index = 0
    non_nan_index = 0

    while real_index < original_length:
        if nan_indexes[real_index]:
            final_series.append(np.NaN)
        else:
            final_series.append(series[non_nan_index])
            non_nan_index += 1
        real_index +=1

    final_series = pd.Series(final_series)
    return final_series


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from .example import MitoWidget, sheet
from .errors import CreationError, EditError
from ._version import __version__, version_info

# Export all the sheet functions
from .sheet_functions.string_functions import *
from .sheet_functions.number_functions import *
from .sheet_functions.date_functions import *
from .sheet_functions.control_functions import *
from .sheet_functions.misc_functions import *

from .nbextension import _jupyter_nbextension_paths



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains all functions that can be used in a sheet that operate on
numbers.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
from functools import reduce
import pandas as pd

from mitosheet.sheet_function_types import as_types, get_series_type_of_first_arg
from mitosheet.sheet_functions.sheet_function_utils import fill_series_list_to_max, fill_series_to_length

@as_types('number_series', remaining=True, ignore_conversion_failures=True)
def AVG(*argv) -> pd.Series:
    """
    {
        "function": "AVG",
        "description": "Returns the numerical mean value of the passed numbers and series.",
        "examples": [
            "AVG(1, 2)",
            "AVG(A, B)",
            "AVG(A, 2)"
        ],
        "syntax": "AVG(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or series to consider when calculating the average."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or series to consider when calculating the average."
            }
        ]
    }
    """

    # Fill sum to the max length of any series
    argv = fill_series_list_to_max(argv)

    arg_sum =  reduce((lambda x, y: x + y), argv) 
    return arg_sum / len(argv)

@as_types(
    'number_series', 
    remaining=True, 
    ignore_conversion_failures=True
)
def MULTIPLY(*argv) -> pd.Series:
    """
    {
        "function": "MULTIPLY",
        "description": "Returns the product of two numbers.",
        "examples": [
            "MULTIPLY(2,3)",
            "MULTIPLY(A,3)"
        ],
        "syntax": "MULTIPLY(factor1, [factor2, ...])",
        "syntax_elements": [{
                "element": "factor1",
                "description": "The first number to multiply."
            },
            {
                "element": "factor2, ... [OPTIONAL]",
                "description": "Additional numbers or series to multiply."
            }
        ]
    }
    """
    # We make sure all the series are the max length, so the * has something at every index
    argv = fill_series_list_to_max(argv)

    return reduce((lambda x, y: x * y), argv) 


@as_types(
    'number_series', 'number_series',
    num_optional=1,
    attempt_output_types=get_series_type_of_first_arg
)
def ROUND(series, decimals=None):
    """
    {
        "function": "ROUND",
        "description": "Rounds a number to a given number of decimals.",
        "examples": [
            "ROUND(1.3)",
            "ROUND(A, 2)"
        ],
        "syntax": "ROUND(value, [decimals])",
        "syntax_elements": [{
                "element": "value",
                "description": "The value or series to round."
            },
            {
                "element": "decimals",
                "description": " The number of decimals to round to. Default is 0."
            }
        ]
    }
    """

    # If no decimals option is passed, round to no decimals
    if decimals is None:
        return series.round()
    
    # Otherwise, fill the decimals to length
    decimals = fill_series_to_length(decimals, series.size)

    return pd.Series(
        [round(num, dec) for num, dec in zip(series, decimals)]
    )


@as_types(
    'number_series'
)
def ABS(series: pd.Series):
    """
    {
        "function": "ABS",
        "description": "Returns the absolute value of the passed number or series.",
        "examples": [
            "ABS(-1.3)",
            "ABS(A)"
        ],
        "syntax": "ABS(value)",
        "syntax_elements": [{
                "element": "value",
                "description": "The value or series to take the absolute value of."
            }
        ]
    }
    """
    return series.abs()


@as_types('number_series', remaining=True, ignore_conversion_failures=True)
def SUM(*argv) -> pd.Series:
    """
    {
        "function": "SUM",
        "description": "Returns the sum of the given numbers and series.",
        "examples": [
            "SUM(10, 11)",
            "SUM(A, B, D, F)",
            "SUM(A, B, D, F)"
        ],
        "syntax": "SUM(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to add together."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to sum."
            }
        ]
    }
    """
    # Fill sum to the max length of any series
    argv = fill_series_list_to_max(argv)

    return reduce((lambda x, y: x + y), argv) 


# TODO: we should see if we can list these automatically!
NUMBER_FUNCTIONS = {
    'ABS': ABS,
    'AVG': AVG,
    'MULTIPLY': MULTIPLY,
    'ROUND': ROUND,
    'SUM': SUM,
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains exports for all functions that can be used in the mitosheet
"""
from mitosheet.sheet_functions.number_functions import NUMBER_FUNCTIONS
from mitosheet.sheet_functions.string_functions import STRING_FUNCTIONS
from mitosheet.sheet_functions.date_functions import DATE_FUNCTIONS
from mitosheet.sheet_functions.control_functions import CONTROL_FUNCTIONS
from mitosheet.sheet_functions.misc_functions import MISC_FUNCTIONS

FUNCTIONS = dict(NUMBER_FUNCTIONS, **STRING_FUNCTIONS, **DATE_FUNCTIONS, **CONTROL_FUNCTIONS, **MISC_FUNCTIONS)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A simple import, which allows you to import dataframes with the given file_names
"""
import pandas as pd
import json
from mitosheet.utils import get_header_renames, make_valid_header

SIMPLE_IMPORT_EVENT = 'simple_import_edit'
SIMPLE_IMPORT_STEP_TYPE = 'simple_import'

SIMPLE_IMPORT_PARAMS = [
    'file_names', # list of strings
]

def file_name_to_df_name(file_name: str):
    # We abuse the fact that all valid mito headers are valid dataframe names...
    # doh.
    return make_valid_header(file_name)

def get_dataframe_names(file_names, existing_df_names):
    """
    Helper function for taking a list of file names and turning them into valid
    names for dataframes.

    NOTE:
    1. If there are duplicates, appends onto the end of them to deduplicate.
    2. Avoids overwriting existing df names.
    """
    new_names_inital = [file_name_to_df_name(file_name) for file_name in file_names]

    final_names = []

    # Keep appending names till we get one that doesn't overlap
    for name in new_names_inital:
        curr_name = name
        count = 0
        while curr_name in final_names or curr_name in existing_df_names:
            curr_name = f'{name}_{count}'
            count += 1
        
        final_names.append(curr_name)
    
    return final_names


def execute_simple_import(
        wsc,
        file_names
    ):
    """
    Imports the files with the given file names into the sheet
    """
    # Create a new step and save the parameters
    wsc._create_and_checkout_new_step(SIMPLE_IMPORT_STEP_TYPE)
    wsc.curr_step['file_names'] = file_names

    column_header_renames = dict()
    for file_name, df_name in zip(file_names, get_dataframe_names(file_names, wsc.curr_step['df_names'])):
        # TODO: catch this error in a good way
        df = pd.read_csv(file_name)
        renames = get_header_renames(df.keys())
        if len(renames) > 0:
            # Save that we did these renames
            column_header_renames[df_name] = renames
            # Actually perform any renames we need to
            df.rename(columns=renames, inplace=True)

        wsc.add_df_to_curr_step(df, df_name=df_name)
    
    # Save the renames that have occured in the step, for transpilation reasons
    wsc.curr_step['column_header_renames'] = column_header_renames


def transpile_simple_import(
        widget_state_container,
        step,
        file_names
    ):
    """
    Transpiles a sort step to Python code. 
    """

    code = ['import pandas as pd']
    for file_name, df_name in zip(file_names, step['df_names'][len(step['df_names']) - len(file_names):]):

        code.append(
            f'{df_name} = pd.read_csv(\'{file_name}\')'
        )
        # If we had to rename columns, mark that as well
        if df_name in step['column_header_renames']:
            renames = step['column_header_renames'][df_name]
            code.append(
                f'{df_name}.rename(columns={json.dumps(renames)}, inplace=True)'
            )

    return code


SIMPLE_IMPORT_STEP = {
    'step_version': 1,
    'event_type': SIMPLE_IMPORT_EVENT,
    'step_type': SIMPLE_IMPORT_STEP_TYPE,
    'params': SIMPLE_IMPORT_PARAMS,
    'execute': execute_simple_import,
    'transpile': transpile_simple_import
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains helpful utility functions for saving and reading
in analyses.
"""

import os
import json

from mitosheet._version import __version__
from mitosheet.steps import STEPS
from mitosheet.profiling import timeit
from mitosheet.mito_analytics import analytics, log_dfs_metadata, static_user_id


# Where all global .mito files are stored
MITO_FOLDER = os.path.expanduser("~/.mito")

# The current version of the saved Mito analysis
# where we save all the analyses for this version
SAVED_ANALYSIS_FOLDER = os.path.join(MITO_FOLDER, 'saved_analyses')


def read_analysis(analysis_name):
    """
    Given an analysis_name, reads the saved analysis in
    ~/.mito/{analysis_name}.json and returns a JSON object
    representing it.
    """
    analysis_path = f'{SAVED_ANALYSIS_FOLDER}/{analysis_name}.json'

    if not os.path.exists(analysis_path):
        return None

    with open(analysis_path) as f:
        try:
            # We try and read the file as JSON
            return json.load(f)
        except: 
            return None

def _get_all_analysis_filenames():
    """
    Returns the names of the files in the SAVED_ANALYSIS_FOLDER
    """
    if not os.path.exists(SAVED_ANALYSIS_FOLDER):
        return []

    file_names = set([
        f for f in os.listdir(SAVED_ANALYSIS_FOLDER) 
        if os.path.isfile(os.path.join(SAVED_ANALYSIS_FOLDER, f))
    ])

    return file_names

def _delete_analyses(analysis_filenames):
    """
    For bulk deleting analysis with file names. 
    """
    for filename in analysis_filenames:
        os.remove(os.path.join(SAVED_ANALYSIS_FOLDER, filename))

def read_saved_analysis_names():
    """
    Reads the names of all the analyses saved by the user.

    Does not return any of the auto-saved analyses!
    """
    if not os.path.exists(SAVED_ANALYSIS_FOLDER):
        return []

    file_names = [
        f for f in os.listdir(SAVED_ANALYSIS_FOLDER) 
        if os.path.isfile(os.path.join(SAVED_ANALYSIS_FOLDER, f))
        and not f.startswith('UUID-')
    ]

    # We make sure they are in alphabetical order!
    file_names.sort()

    return [
        file_name[:-5] for file_name in file_names 
        if file_name.endswith('.json')
    ]

@timeit
def saved_analysis_names_json():
    return json.dumps(read_saved_analysis_names())

def make_steps_json_obj(steps):
    """
    Given a steps dictonary from a widget_state_container, puts the steps
    into a format that can be saved and recreated. Necessary for saving an
    analysis to a file!
    """
    steps_json_obj = dict()
    for step_id, step in enumerate(steps):
        step_type = step['step_type']
        
        for new_step in STEPS:
            if step_type == new_step['step_type']:

                # Save the step type
                step_summary = {
                    'step_version': new_step['step_version'],
                    'step_type': step_type,
                }
                # As well as all of the parameters for the step
                step_summary.update({key: value for key, value in step.items() if key in new_step['params']})

                steps_json_obj[step_id] = step_summary

    return steps_json_obj

@timeit
def write_analysis(widget_state_container, analysis_name=None):
    """
    Writes the analysis saved in widget_state_container to
    ~/.mito/{analysis_name}. If analysis_name is none, gets the temporary
    name from the widget_state_container.

    NOTE: as the written analysis is from the widget_state_container,
    we assume that the analysis is valid when written and read back in!
    """

    if not os.path.exists(MITO_FOLDER):
        os.mkdir(MITO_FOLDER)

    if not os.path.exists(SAVED_ANALYSIS_FOLDER):
        os.mkdir(SAVED_ANALYSIS_FOLDER)

    if analysis_name is None:
        analysis_name = widget_state_container.analysis_name

    analysis_path = f'{SAVED_ANALYSIS_FOLDER}/{analysis_name}.json'

    with open(analysis_path, 'w+') as f:
        steps_json_obj = make_steps_json_obj(widget_state_container.steps)
        saved_analysis = {
            'version': __version__,
            'steps': steps_json_obj
        }

        f.write(json.dumps(saved_analysis))
    
    # We then log that we saved the analysis, if it is not a temporary save
    if not analysis_name.startswith('UUID'):
        # Log metadata about these dataframes
        log_dfs_metadata(widget_state_container.dfs)

        # And then 
        step_types = [step['step_type'] for step in widget_state_container.steps]
        analytics.track(static_user_id, 'write_analysis_log_event', {
            'step_types': step_types
        })





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
See mito/mitosheet/steps/README.md for more information about 
how to add a step!

NOTE: this new step refactor is a WIP, and will not be finished
until we remove all the manual step casing that occurs throughout the 
codebase. This is an incremental process that will take time!
"""

from mitosheet.steps.group import GROUP_STEP
from mitosheet.steps.filter import FILTER_STEP
from mitosheet.steps.sort import SORT_STEP
from mitosheet.steps.reorder_column import REORDER_COLUMN_STEP
from mitosheet.steps.add_column import ADD_COLUMN_STEP
from mitosheet.steps.set_column_formula import SET_COLUMN_FORMULA_STEP
from mitosheet.steps.merge import MERGE_STEP
from mitosheet.steps.column_delete import DELETE_COLUMN_STEP
from mitosheet.steps.column_rename import RENAME_COLUMN_STEP
from mitosheet.steps.import_steps.simple_import import SIMPLE_IMPORT_STEP
from mitosheet.steps.import_steps.raw_python_import import RAW_PYTHON_IMPORT_STEP

# All steps must be listed in this variable.
STEPS = [
    GROUP_STEP,
    REORDER_COLUMN_STEP,
    FILTER_STEP,
    SORT_STEP,
    ADD_COLUMN_STEP,
    SET_COLUMN_FORMULA_STEP,
    MERGE_STEP,
    DELETE_COLUMN_STEP,
    RENAME_COLUMN_STEP,
    SIMPLE_IMPORT_STEP,
    RAW_PYTHON_IMPORT_STEP
]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains all functions that are useful for control flow. For now, this
is just IF statements.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
import pandas as pd
from functools import reduce

from mitosheet.sheet_function_types import as_types, get_series_type_of_first_arg, MITO_TYPE_CONVERSION_MAP
from mitosheet.sheet_functions.sheet_function_utils import fill_series_list_to_max, fill_series_to_length

@as_types(
    'boolean_series', 
    remaining=True
)
def AND(*argv) -> pd.Series:
    """
    {
        "function": "AND",
        "description": "Returns True if all of the provided arguments are True, and False if any of the provided arguments are False.",
        "examples": [
            "AND(True, False)",
            "AND(Nums > 100, Nums < 200)",
            "AND(Pay > 10, Pay < 20, Status == 'active')"
        ],
        "syntax": "AND(boolean_condition1, [boolean_condition2, ...])",
        "syntax_elements": [{
                "element": "boolean_condition1",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            },
            {
                "element": "boolean_condition2 ... [OPTIONAL]",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            }
        ]
    }
    """
    argv = fill_series_list_to_max(argv)

    return reduce((lambda x, y: x & y), argv)


@as_types(
    'boolean_series', 
    'series',
    'series'
)
def IF(condition, true_series, false_series) -> pd.Series:
    """
    {
        "function": "IF",
        "description": "Returns one value if the condition is True. Returns the other value if the conditon is False.",
        "examples": [
            "IF(Status == 'success', 1, 0)",
            "IF(Nums > 100, 100, Nums)",
            "IF(AND(Grade >= .6, Status == 'active'), 'pass', 'fail')"
        ],
        "syntax": "IF(boolean_condition, value_if_true, value_if_false)",
        "syntax_elements": [{
                "element": "boolean_condition",
                "description": "An expression or series that returns True or False values. Valid conditions for comparison include ==, !=, >, <, >=, <=."
            },
            {
                "element": "value_if_true",
                "description": "The value the function returns if condition is True."
            },
            {
                "element": "value_if_false",
                "description": "The value the function returns if condition is False."
            }
        ]
    }
    """
    # NOTE: condition should just be a boolean series. true_series, false_series can be series
    # of any type!

    true_series = fill_series_to_length(true_series, len(condition))
    false_series = fill_series_to_length(false_series, len(condition))


    return pd.Series(
        data=[true_series[i] if c else false_series[i] for i, c in enumerate(condition)]
    )

@as_types(
    'boolean_series', 
    remaining=True
)
def OR(*argv) -> pd.Series:
    """
    {
        "function": "OR",
        "description": "Returns True if any of the provided arguments are True, and False if all of the provided arguments are False.",
        "examples": [
            "OR(True, False)",
            "OR(Status == 'success', Status == 'pass', Status == 'passed')"
        ],
        "syntax": "OR(boolean_condition1, [boolean_condition2, ...])",
        "syntax_elements": [{
                "element": "boolean_condition1",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            },
            {
                "element": "boolean_condition2 ... [OPTIONAL]",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            }
        ]
    }
    """

    argv = fill_series_list_to_max(argv)
    return reduce((lambda x, y: x | y), argv)



# TODO: we should see if we can list these automatically!
CONTROL_FUNCTIONS = {
    'AND': AND,
    'IF': IF,
    'OR': OR,
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A add_column step, which allows you to add a column to 
a dataframe. 
"""

from mitosheet.errors import (
    make_column_exists_error,
    make_no_sheet_error
)

ADD_COLUMN_EVENT = 'add_column_edit'
ADD_COLUMN_STEP_TYPE = 'add_column'

ADD_COLUMN_PARAMS = [
    'sheet_index', # int
    'column_header', # the new column to create
]

def execute_add_column_step(
        wsc,
        sheet_index,
        column_header
    ):
    """
    The function responsible for updating the widget state container
    with a new add_column step.
    """
    
    # if the sheet doesn't exist, throw an error
    if not wsc.does_sheet_index_exist(sheet_index):
        raise make_no_sheet_error(sheet_index)

    if column_header in wsc.curr_step['column_metatype'][sheet_index]:
        raise make_column_exists_error(column_header)

    # We add a new step with the added column
    wsc._create_and_checkout_new_step(ADD_COLUMN_STEP_TYPE)
    # Save the params
    wsc.curr_step['sheet_index'] = sheet_index
    wsc.curr_step['column_header'] = column_header

    # Update the state variables
    wsc.curr_step['column_metatype'][sheet_index][column_header] = 'formula'
    wsc.curr_step['column_type'][sheet_index][column_header] = 'number'
    wsc.curr_step['column_spreadsheet_code'][sheet_index][column_header] = '=0'
    wsc.curr_step['column_python_code'][sheet_index][column_header] = f'df[\'{column_header}\'] = 0'
    wsc.curr_step['column_evaluation_graph'][sheet_index][column_header] = set()
    wsc.curr_step['column_filters'][sheet_index][column_header] = {'operator': 'And', 'filters': []}

    # Update the dataframe
    wsc.curr_step['dfs'][sheet_index][column_header] = 0


def transpile_add_column_step(
        widget_state_container,
        step,
        sheet_index,
        column_header
    ):
    """
    Transpiles an add column step to python code!
    """

    return [
        f'{widget_state_container.curr_step["df_names"][sheet_index]}[\'{column_header}\'] = 0'
    ]


ADD_COLUMN_STEP = {
    'step_version': 1,
    'event_type': ADD_COLUMN_EVENT,
    'step_type': ADD_COLUMN_STEP_TYPE,
    'params': ADD_COLUMN_PARAMS,
    'execute': execute_add_column_step,
    'transpile': transpile_add_column_step
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
NOTE: this step should not be included in the export in __init__.py, as it is 
a special sort of step that does not get saved in the step list!

An initial rename, which is a step that only occurs at the start of an analysis (and
does not error if replayed when already run), but is responsible for renaming invalid
column headers to headers mito can handle.

NOTE: this step is not the same as a column rename for a variety of reasons, including:
1. With many column headers getting renamed, indivigual column rename steps generate too much code. 
   Though this could be optimized out, doing so it tricky with our current approach of inserting
   formula steps in-between all steps.
2. This step is automatically generated at the start of every analysis it needs to be generated at,
   and so needs to handle being replayed differently than column renames (because it should never
   error if it has been run before, because it _always_ has run before when replayed)!
"""
import json
from mitosheet.utils import get_header_renames, is_valid_header, make_valid_header

INITIAL_RENAME_EVENT = 'initial_rename_edit' # NOTE: this should never be generated by the frontend!
INITIAL_RENAME_STEP_TYPE = 'initial_rename'

# It has no parameters, other than the dataframes in the wsc!
INITIAL_RENAME_PARAMS = []

def execute_initial_rename_step(dfs):
    """
    The function responsible for updating the widget state container
    with an initial rename step. This is not logged as step in the step list
    and just implicitly occurs at the start of every analysis!

    """
    for sheet_index, df in enumerate(dfs):
        renames = dict()
        for column_header in df.keys():
            if not is_valid_header(column_header):
                valid_header = make_valid_header(column_header)
                renames[column_header] = valid_header
        
        df.rename(columns=renames, inplace=True)

    return dfs

def transpile_initial_rename_step(
        widget_state_container
    ):
    """
    Transpiles an initial rename to Python code! May be empty if there is nothing to rename.

    NOTE: we still have a df naming problem, and so in this case, we simply use the names of the 
    dataframes as they are pased in the initial step, which has them unchanged, as they originially
    were.
    """
    rename_lines = []
    for sheet_index, column_headers in enumerate(widget_state_container.original_df_keys):
        renames = get_header_renames(column_headers)
        
        if len(renames) == 0:
            continue

        rename_lines.append(
            f'{widget_state_container.curr_step["df_names"][sheet_index]}.rename(columns={json.dumps(renames)}, inplace=True)'
        )

    return rename_lines


"""
This object wraps all the information for the initial rename step
"""
INITIAL_RENAME_STEP = {
    'event_type': INITIAL_RENAME_EVENT,
    'step_type': INITIAL_RENAME_STEP_TYPE,
    'params': INITIAL_RENAME_PARAMS,
    'execute': execute_initial_rename_step,
    'transpile': transpile_initial_rename_step
}








#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains the exports for update events, which are events that manipulate
the steps in some way, but do not _just generate a step_ (see ../steps for
more information about steps).

Examples of this sort of event are:
- Lots of steps being replayed from the front-end.
- A user undoing an existing step. 
"""

from mitosheet.updates.undo import UNDO_UPDATE


# All steps must be listed in this variable.
UPDATES = [
    UNDO_UPDATE
]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Initial steps are steps that:
1. Are run at the start of every Mito analysis, (e.g. right after
   a user calls mitosheet.sheet).
2. Require no user input to run - and are just always run - and so
   we don't necessarily need to save them when we write out an 
   analysis.

Currently, there are two initial steps:
1. Renaming column headers to be a valid Mito format. 
2. Creating the initial "step."
"""

from mitosheet.steps.initial_steps.initalize import INITIALIZE_STEP
from mitosheet.steps.initial_steps.initial_rename import INITIAL_RENAME_STEP

INITIAL_STEPS = [
    INITIALIZE_STEP,
    INITIAL_RENAME_STEP
]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
An initalize step, which is responsible for creating the _very first_
step in any mito analysis. It should transpile to nothing.
"""

from collections import OrderedDict
from mitosheet.utils import get_column_filter_type

INITIALIZE_EVENT = None
INITIALIZE_STEP_TYPE = 'initialize'

INITIALIZE_PARAMS = []

def execute_initialize_step(wsc, dfs):
    """
    The function responsible for updating the widget state container
    with a new initialize step.

    If it fails part of the way through, deletes the new initialize step entirely.
    """

    if len(wsc.steps) > 0:
        raise Exception("This step should only be applied at the start of an analysis")

    wsc.steps = [{
        'step_type': INITIALIZE_STEP_TYPE, # the first step is a unique step type that is just the first step!
        # The column_metatype is if it stores formulas or values
        'column_metatype': [{key: 'value' for key in df.keys()} for df in dfs], 
        # The column_type is the type of the series in this column 
        'column_type': [{key: get_column_filter_type(df[key]) for key in df.keys()} for df in dfs],
        # We make column_spreadsheet_code an ordered dictonary to preserve the order the formulas
        # are inserted, which in turn makes sure when we save + rerun an analysis, it's recreated
        # in the correct order (and thus the column order is preserved).
        'column_spreadsheet_code': [OrderedDict({key: '' for key in df.keys()}) for df in dfs],
        # We also keep track of a list of the added columns in this step, so we can make
        # sure columns get defined in the correct order in the transpiled code
        'added_columns': [[] for df in dfs],
        'column_python_code': [{key: '' for key in df.keys()} for df in dfs],
        'column_evaluation_graph': [{key: set() for key in df.keys()} for df in dfs],
        'column_filters': [{key: {'operator': 'And', 'filters': []} for key in df.keys()} for df in dfs],
        'dfs': dfs,
        # The df_names are composed of two parts:
        # 1. The names of the variables passed into the mitosheet.sheet call (which don't change over time).
        # 2. The names of the dataframes that were created during the analysis (e.g. by a merge).
        # Until we get them from the frontend as an update_event, we default them to df1, df2, ...
        'df_names': [f'df{i + 1}' for i in range(len(dfs))] 
    }]


def transpile_initialize_step(widget_state_container, step):
    """
    Transpiles a initialize step to python code!
    """
    return []

"""
This object wraps all the information
that is needed for a initialize step!
"""
INITIALIZE_STEP = {
    'event_type': INITIALIZE_EVENT,
    'step_type': INITIALIZE_STEP_TYPE,
    'params': INITIALIZE_PARAMS,
    'execute': execute_initialize_step,
    'transpile': transpile_initialize_step
}








#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import pytest
import pandas as pd

from mitosheet.transpiler.transpile import transpile
from mitosheet.tests.test_utils import create_mito_wrapper, create_mito_wrapper_dfs


def test_transpile_single_column():
    mito = create_mito_wrapper(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == ['# Step 1', 'df1[\'B\'] = df1[\'A\']']


def test_transpile_multiple_columns_no_relationship():
    mito = create_mito_wrapper(['abc'])
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')
    code_container = transpile(mito.mito_widget.widget_state_container)
    
    assert len([line for line in code_container['code'] if 'df1[\'B\'] = 0' in line]) > 0
    assert len([line for line in code_container['code'] if 'df1[\'C\'] = 0' in line]) > 0

def test_transpile_columns_in_each_sheet():
    mito = create_mito_wrapper(['abc'], sheet_two_A_data=['abc'])
    mito.add_column(0, 'B')
    mito.add_column(1, 'B')
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert len([line for line in code_container['code'] if 'df1[\'B\'] = 0' in line]) > 0
    assert len([line for line in code_container['code'] if 'df2[\'B\'] = 0' in line]) > 0

def test_transpile_multiple_columns_linear():
    mito = create_mito_wrapper(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)
    mito.set_formula('=B', 0, 'C', add_column=True)
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1', 
        'df1[\'B\'] = df1[\'A\']', 
        '# Step 2', 
        'df1[\'C\'] = df1[\'B\']'
    ]

COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet(column_header):
    mito = create_mito_wrapper(['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == ['# Step 1', f'df1[\'{column_header}\'] = df1[\'A\']']


COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet_multi_sheet(column_header):
    mito = create_mito_wrapper(['abc'], sheet_two_A_data=['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)
    mito.set_formula('=A', 1, column_header, add_column=True)
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1', 
        f'df1[\'{column_header}\'] = df1[\'A\']', 
        '# Step 2', 
        f'df2[\'{column_header}\'] = df2[\'A\']'
    ]

def test_preserves_order_columns():
    mito = create_mito_wrapper(['abc'])
    # Topological sort will currently display this in C, B order
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')
    code_container = transpile(mito.mito_widget.widget_state_container)
    # Note: this optimizes out unnecessary code
    assert code_container['code'] == [
        '# Step 1', 
        "df1['B'] = 0",
        '# Step 2', 
        "df1['C'] = 0"
    ]

def test_transpile_merge_all_columns():
    df_one = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df_two = pd.DataFrame(data={'A': [1], 'D': [100], 'E': [10]})
    mito = create_mito_wrapper_dfs(df_one, df_two)

    mito.merge_sheets(0, 'A', list(df_one.keys()), 1, 'A', list(df_two.keys()))
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1', 
        'temp_df = df2.drop_duplicates(subset=\'A\')', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])'
    ]

def test_transpile_merge_some_all_columns():
    df_one = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df_two = pd.DataFrame(data={'A': [1], 'D': [100], 'E': [10]})
    mito = create_mito_wrapper_dfs(df_one, df_two)

    mito.merge_sheets(0, 'A', ['A', 'B'], 1, 'A', ['A', 'D'])
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1', 
        "temp_df = df2.drop_duplicates(subset='A')", 
        "df1_tmp = df1.drop(['C'], axis=1)", 
        "df2_tmp = temp_df.drop(['E'], axis=1)", 
        "df3 = df1_tmp.merge(df2_tmp, left_on=['A'], right_on=['A'], how='left', suffixes=['_df1', '_df2'])"
    ]

def test_transpile_merge_all_columns_than_some_columns():
    df_one = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df_two = pd.DataFrame(data={'A': [1], 'D': [100], 'E': [10]})
    mito = create_mito_wrapper_dfs(df_one, df_two)

    mito.merge_sheets(0, 'A', list(df_one.keys()), 1, 'A', list(df_two.keys()))    
    mito.merge_sheets(0, 'A', ['A', 'B'], 1, 'A', ['A', 'D'])

    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1', 
        'temp_df = df2.drop_duplicates(subset=\'A\')', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '# Step 2', 
        "temp_df = df2.drop_duplicates(subset='A')", 
        "df1_tmp = df1.drop(['C'], axis=1)", 
        "df2_tmp = temp_df.drop(['E'], axis=1)", 
        "df4 = df1_tmp.merge(df2_tmp, left_on=['A'], right_on=['A'], how='left', suffixes=['_df1', '_df2'])"
    ]

def test_transpile_delete_column():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1)
    mito.delete_column(0, 'C')

    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1', 
        'df1.drop(\'C\', axis=1, inplace=True)'
    ]


# TESTING OPTIMIZATION

def test_removes_unedited_formulas_for_unedited_sheets():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1, df2)

    mito.set_formula('=C', 0, 'D', add_column=True)
    mito.set_formula('=C', 1, 'D', add_column=True)

    mito.merge_sheets(0, 'A', ['A', 'B', 'C', 'D'], 1, 'A', ['A', 'B', 'C', 'D'])

    mito.set_formula('=C + 1', 1, 'D', add_column=True)

    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1', 
        'df1[\'D\'] = df1[\'C\']',
        '# Step 2', 
        'df2[\'D\'] = df2[\'C\']',
        '# Step 3', 
        'temp_df = df2.drop_duplicates(subset=\'A\')', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '# Step 4', 
        'df2[\'D\'] = df2[\'C\'] + 1',
    ]


def test_mulitple_merges_no_formula_steps():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.merge_sheets(0, 'A', ['A', 'B', 'C'], 1, 'A', ['A', 'B', 'C'])
    mito.merge_sheets(0, 'A', ['A', 'B', 'C'], 1, 'A', ['A', 'B', 'C'])
    mito.merge_sheets(0, 'A', ['A', 'B', 'C'], 1, 'A', ['A', 'B', 'C'])


    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1', 
        'temp_df = df2.drop_duplicates(subset=\'A\')', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '# Step 2', 
        'temp_df = df2.drop_duplicates(subset=\'A\')', 
        'df4 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '# Step 3', 
        'temp_df = df2.drop_duplicates(subset=\'A\')', 
        'df5 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])'
    ]

def test_optimization_with_other_edits():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.add_column(0, 'D')
    mito.set_formula('=A', 0, 'D')
    mito.merge_sheets(0, 'A', ['A', 'B', 'C', 'D'], 1, 'A', ['A', 'B', 'C'])
    mito.add_column(0, 'AAA')
    mito.delete_column(0, 'AAA')
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1', 
        'df1[\'D\'] = df1[\'A\']',
        '# Step 2', 
        'temp_df = df2.drop_duplicates(subset=\'A\')', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '# Step 3',
        'df1[\'AAA\'] = 0',
        '# Step 4', 
        'df1.drop(\'AAA\', axis=1, inplace=True)'
    ]

TAB = '    '
NEWLINE_TAB = f'\n{TAB}'
flatten_comment = '# Flatten the column headers'
flatten_code = f'pivot_table.columns = [{NEWLINE_TAB}\'_\'.join([str(c) for c in col]).strip() if isinstance(col, tuple) else col{NEWLINE_TAB}for col in pivot_table.columns.values\n]'
FLATTEN_LINE = f'{flatten_comment}\n{flatten_code}'

def test_group_transpiles_single_rows():
    df1 = pd.DataFrame(data={'Name': ['Nate', 'Nate'], 'Height': [4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.group_sheet(
        0, ['Name'], [], {'Height': 'sum'}
    )
    code_container = transpile(mito.mito_widget.widget_state_container)

    first_line = """# Pivot the data
pivot_table = df1.pivot_table(
    index=['Name'],
    values=['Height'],
    aggfunc={'Height': 'sum'}
)"""

    print(code_container)
    second_Line = """# Reset the column name and the indexes
df2 = pivot_table.rename_axis(None, axis=1).reset_index()"""
    assert code_container['code'] == [
        '# Step 1',
        first_line,
        second_Line
    ]

def test_group_transpiles_single_columns():
    df1 = pd.DataFrame(data={'Name': ['Nate', 'Nate'], 'Height': [4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.group_sheet(
        0, [], ['Name'], {'Height': 'sum'}
    )
    code_container = transpile(mito.mito_widget.widget_state_container)

    first_line = """# Pivot the data
pivot_table = df1.pivot_table(
    columns=['Name'],
    values=['Height'],
    aggfunc={'Height': 'sum'}
)"""

    second_Line = """# Reset the column name and the indexes
df2 = pivot_table.rename_axis(None, axis=1).reset_index()"""
    assert code_container['code'] == [
        '# Step 1',
        first_line,
        FLATTEN_LINE,
        second_Line
    ]


def test_group_transpiles_single_column_overlap_group_by_values():
    df1 = pd.DataFrame(data={'Name': ['Nate', 'Nate'], 'Height': [4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.group_sheet(
        0, ['Name'], [], {'Name': 'count'}
    )
    mito.group_sheet(
        0, ['Name'], [], {'Name': 'count'}
    )
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'groupby_obj = df1.groupby([\'Name\'], as_index=False)',
        'df2 = groupby_obj.size()',
        '# Step 2',
        'groupby_obj = df1.groupby([\'Name\'], as_index=False)',
        'df3 = groupby_obj.size()',
    ]

def test_group_leaves_early_with_no_agg_columns():
    df1 = pd.DataFrame(data={'Name': ['Nate', 'Nate'], 'Height': [4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.group_sheet(
        0, ['Name'], [], {}
    )

    first_line = """# Pivot the data
pivot_table = df1.pivot_table(
    index=['Name'],
)"""

    second_line = """# Reset the column name and the indexes
df2 = pivot_table.rename_axis(None, axis=1).reset_index()"""
    
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        first_line,
        second_line
    ]

def test_group_transpiles_group_by_mulitple_columns():
    df1 = pd.DataFrame(data={'First_Name': ['Nate', 'Nate'], 'Last_Name': ['Rush', 'Jack'], 'Height': [4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.group_sheet(
        0, ['First_Name', 'Last_Name'], [], {'Height': 'sum'}
    )

    first_line = """# Pivot the data
pivot_table = df1.pivot_table(
    columns=['First_Name', 'Last_Name'],
    values=['Height'],
    aggfunc={'Height': 'sum'}
)"""

    second_line = """# Flatten the column headers
pivot_table.columns = [
    '_'.join([str(c) for c in col]).strip() if isinstance(col, tuple) else col
    for col in pivot_table.columns.values
]"""

    third_line = """# Reset the column name and the indexes
df2 = pivot_table.rename_axis(None, axis=1).reset_index()"""

    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        first_line, 
        second_line,
        third_line
    ]


def test_group_transpiles_group_by_mulitple_columns():
    df1 = pd.DataFrame(data={'First_Name': ['Nate', 'Nate'], 'Last_Name': ['Rush', 'Jack'], 'Height': [4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.group_sheet(
        0, ['First_Name', 'Last_Name'], [], {'Height': 'sum'}
    )

    first_line = """# Pivot the data
pivot_table = df1.pivot_table(
    index=['First_Name', 'Last_Name'],
    values=['Height'],
    aggfunc={'Height': 'sum'}
)"""

    second_line = """# Reset the column name and the indexes
df2 = pivot_table.rename_axis(None, axis=1).reset_index()"""

    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        first_line, 
        second_line
    ]

def test_group_transpiles_group_by_mulitple_columns_agg():
    df1 = pd.DataFrame(data={'First_Name': ['Nate', 'Nate'], 'Last_Name': ['Rush', 'Jack'], 'Height': [4, 5], 'Weight': [5, 6]})
    mito = create_mito_wrapper_dfs(df1)
    mito.group_sheet(
        0, ['First_Name', 'Last_Name'], [], {'Height': 'sum', 'Weight': 'mean'}
    )

    first_line = """# Pivot the data
pivot_table = df1.pivot_table(
    index=['First_Name', 'Last_Name'],
    values=['Height', 'Weight'],
    aggfunc={'Height': 'sum', 'Weight': 'mean'}
)"""

    second_line = """# Reset the column name and the indexes
df2 = pivot_table.rename_axis(None, axis=1).reset_index()"""

    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        first_line,
        second_line
    ]

def test_group_transpiles_mulitple_sheets():
    df1 = pd.DataFrame(data={'First_Name': ['Nate', 'Nate'], 'Last_Name': ['Rush', 'Jack'], 'Height': [4, 5], 'Weight': [5, 6]})
    df2 = pd.DataFrame(data={'First_Name': ['Nate', 'Nate'], 'Last_Name': ['Rush', 'Jack'], 'Height': [4, 5], 'Weight': [5, 6]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.group_sheet(
        0, ['First_Name', 'Last_Name'], [], {'Height': 'sum', 'Weight': 'mean'}
    )
    mito.group_sheet(
        1, ['First_Name', 'Last_Name'], [], {'Height': 'sum', 'Weight': 'mean'}
    )

    pivot_line1 = """# Pivot the data
pivot_table = df1.pivot_table(
    index=['First_Name', 'Last_Name'],
    values=['Height', 'Weight'],
    aggfunc={'Height': 'sum', 'Weight': 'mean'}
)"""
    pivot_line2 = """# Pivot the data
pivot_table = df2.pivot_table(
    index=['First_Name', 'Last_Name'],
    values=['Height', 'Weight'],
    aggfunc={'Height': 'sum', 'Weight': 'mean'}
)"""

    reset_line1 = """# Reset the column name and the indexes
df3 = pivot_table.rename_axis(None, axis=1).reset_index()"""
    reset_line2 = """# Reset the column name and the indexes
df4 = pivot_table.rename_axis(None, axis=1).reset_index()"""


    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        pivot_line1,
        reset_line1,
        '# Step 2',
        pivot_line2,
        reset_line2,
    ]


def test_transpile_initial_rename_mulitple_columns():
    df1 = pd.DataFrame(data={'First Name': ['Nate', 'Nate'], 'Last Name': ['Rush', 'Jack']})
    mito = create_mito_wrapper_dfs(df1)
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1 (rename headers to make them work with Mito)',
        'df1.rename(columns={\"First Name\": \"First_Name\", \"Last Name\": \"Last_Name\"}, inplace=True)',
    ]

def test_transpile_initial_rename_multiple_dfs():
    df1 = pd.DataFrame(data={'First Name': ['Nate', 'Nate'], 'Last_Name': ['Rush', 'Jack']})
    df2 = pd.DataFrame(data={'First_Name': ['Nate', 'Nate'], 'Last Name': ['Rush', 'Jack']})
    mito = create_mito_wrapper_dfs(df1, df2)
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1 (rename headers to make them work with Mito)',
        'df1.rename(columns={\"First Name\": \"First_Name\"}, inplace=True)',
        'df2.rename(columns={\"Last Name\": \"Last_Name\"}, inplace=True)',
    ]

def test_transpile_filter():
    df1 = pd.DataFrame(data={'name': ['Nate', 'Jake'], 'Last_Name': ['Rush', 'Jack']})
    mito = create_mito_wrapper_dfs(df1)
    mito.filter(0, 'name', 'And', 'string', 'contains', 'Nate')
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1 = df1[df1[\'name\'].str.contains(\'Nate\', na=False)]',
        'df1 = df1.reset_index(drop=True)',
    ]


def test_transpile_filter_string_does_not_contain():
    df1 = pd.DataFrame(data={'name': ['Nate', 'Jake'], 'Last_Name': ['Rush', 'Jack']})
    mito = create_mito_wrapper_dfs(df1)
    mito.filter(0, 'name', 'And', 'string', 'string_does_not_contain', 'Nate')
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1 = df1[~df1[\'name\'].str.contains(\'Nate\', na=False)]',
        'df1 = df1.reset_index(drop=True)',
    ]

def test_transpile_date_filter():
    df = pd.DataFrame(data={'A': pd.to_datetime(pd.Series(data=[
        '12-2-2020', 
        '12-3-2020'
    ]))})
    mito = create_mito_wrapper_dfs(df)
    mito.filter(0, 'A', 'And', 'datetime', 'datetime_greater', '12-2-2020')
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1 = df1[df1[\'A\'] > pd.to_datetime(\'12-2-2020\')]',
        'df1 = df1.reset_index(drop=True)',
    ]

def test_transpile_double_filter_and():
    df1 = pd.DataFrame(data={'name': ['Nate', 'Jake'], 'Last_Name': ['Rush', 'Jack']})
    mito = create_mito_wrapper_dfs(df1)
    mito.filters(0, 'name', 'And', [{'type': 'string', 'condition': 'contains', 'value': 'e'}, {'type': 'string', 'condition': 'string_exactly', 'value': 'Nate'}])
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1 = df1[(df1[\'name\'].str.contains(\'e\', na=False)) & (df1[\'name\'] == \'Nate\')]',
        'df1 = df1.reset_index(drop=True)',
    ]

def test_transpile_double_filter_or():
    df1 = pd.DataFrame(data={'name': ['Nate', 'Jake'], 'Last_Name': ['Rush', 'Jack']})
    mito = create_mito_wrapper_dfs(df1)
    mito.filters(0, 'name', 'Or', [{'type': 'string', 'condition': 'contains', 'value': 'e'}, {'type': 'string', 'condition': 'string_exactly', 'value': 'Nate'}])
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1 = df1[(df1[\'name\'].str.contains(\'e\', na=False)) | (df1[\'name\'] == \'Nate\')]',
        'df1 = df1.reset_index(drop=True)',
    ]

def test_transpile_triple_filter():
    df1 = pd.DataFrame(data={'name': ['Nate', 'Jake'], 'Last_Name': ['Rush', 'Jack']})
    mito = create_mito_wrapper_dfs(df1)
    mito.filters(0, 'name', 'Or', [{'type': 'string', 'condition': 'contains', 'value': 'e'}, {'type': 'string', 'condition': 'contains', 'value': 'a'}, {'type': 'string', 'condition': 'string_exactly', 'value': 'Nate'}])
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1 = df1[(df1[\'name\'].str.contains(\'e\', na=False)) | (df1[\'name\'].str.contains(\'a\', na=False)) | (df1[\'name\'] == \'Nate\')]',
        'df1 = df1.reset_index(drop=True)',
    ]
    
def test_transpile_reorder_column():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper_dfs(df1)
    mito.reorder_column(0, 'A', 1)
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]'
    ]

def test_transpile_two_column_reorders():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper_dfs(df1)
    mito.reorder_column(0, 'A', 1)
    mito.reorder_column(0, 'B', 1)
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
        '# Step 2',
        'df1_columns = [col for col in df1.columns if col != \'B\']',
        'df1_columns.insert(1, \'B\')',
        'df1 = df1[df1_columns]'
    ]

def test_transpile_reorder_column_invalid():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper_dfs(df1)
    mito.reorder_column(0, 'A', 5)
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
    ]

def test_transpile_sort_ascending_valid():
    df1 = pd.DataFrame(data={'A': [1, 2, 3, 4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.sort(0, 'A', 'ascending')
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1 = df1.sort_values(by=\'A\', ascending=True, na_position=\'first\')',
        'df1 = df1.reset_index(drop=True)',
    ]

def test_transpile_sort_descending_valid():
    df1 = pd.DataFrame(data={'A': [1, 2, 3, 4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.sort(0, 'A', 'descending')
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1 = df1.sort_values(by=\'A\', ascending=False, na_position=\'first\')',
        'df1 = df1.reset_index(drop=True)',
    ]

def test_transpile_sort_ascending_then_descending_valid():
    df1 = pd.DataFrame(data={'A': [1, 2, 3, 4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.sort(0, 'A', 'ascending')
    mito.sort(0, 'A', 'descending')

    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'df1 = df1.sort_values(by=\'A\', ascending=True, na_position=\'first\')',
        'df1 = df1.reset_index(drop=True)',
        '# Step 2',
        'df1 = df1.sort_values(by=\'A\', ascending=False, na_position=\'first\')',
        'df1 = df1.reset_index(drop=True)'
    ]

def test_transpile_merge_then_sort():
    df1 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Number': [123, 1]})
    df2 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Sign': ['Gemini', "Tarus"]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.merge_sheets(0, 'Name', list(df1.keys()), 1, 'Name', list(df2.keys()))
    mito.sort(2, 'Number', 'ascending')
    code_container = transpile(mito.mito_widget.widget_state_container)
    assert code_container['code'] == [
        '# Step 1',
        'temp_df = df2.drop_duplicates(subset=\'Name\')',
        'df3 = df1.merge(temp_df, left_on=[\'Name\'], right_on=[\'Name\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '# Step 2',
        'df3 = df3.sort_values(by=\'Number\', ascending=True, na_position=\'first\')',
        'df3 = df3.reset_index(drop=True)'
    ]



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from typing import List
import uuid
import sys
import pandas as pd
import json
from copy import deepcopy, copy

from mitosheet.mito_analytics import analytics, log_dfs_metadata, static_user_id
from mitosheet.steps import STEPS
from mitosheet.updates import UPDATES
from mitosheet.steps.initial_steps.initalize import execute_initialize_step
from mitosheet.steps.initial_steps.initial_rename import execute_initial_rename_step
from mitosheet.sheet_functions import FUNCTIONS
from mitosheet.utils import dfs_to_json, get_column_filter_type
from mitosheet.transpiler.transpile import transpile
from mitosheet.save_utils import write_analysis
from mitosheet.errors import (
    make_execution_error
)
from mitosheet.profiling import timeit


class WidgetStateContainer():
    """
    Holds the state of the steps, which represents operations that
    have been performed on the input dataframes. 
    """

    def __init__(self, dfs: pd.DataFrame):
        # Just in case they are a tuple, make them a list - as it's easier to operate with.
        # We also make a copy so we don't modify the original dataframes.
        dfs = deepcopy(list(dfs))

        # Log metadata about these dataframes
        log_dfs_metadata(dfs)

        # We also save a copy of the initial dataframe's keys, for easy access
        self.original_df_keys = [df.keys() for df in dfs]

        # For now, we just randomly generate analysis names. 
        # We append a UUID to note that this is not an analysis the user has saved.
        self.analysis_name = 'UUID-' + str(uuid.uuid4())

        # We generate an initial rename step, which handles any issues with invalid headers
        # by converting them all to valid headers - which occurs before any formula step.
        dfs = execute_initial_rename_step(dfs)

        # Then we initialize the first initalize step
        self.steps = []
        execute_initialize_step(self, dfs)

    @property
    @timeit
    def curr_step_id(self):
        """
        The ID of the current step
        """
        return len(self.steps) - 1

    @property
    @timeit
    def curr_step(self):
        """
        Returns the current step object as a property of the object,
        so reference it with self.curr_step
        """
        return self.steps[self.curr_step_id]

    @property
    @timeit
    def num_sheets(self):
        """
        Duh. :)
        """
        return len(self.steps[self.curr_step_id]['dfs'])

    @property
    @timeit
    def dfs(self) -> List[pd.DataFrame]:
        return self.steps[self.curr_step_id]['dfs']

    @property
    @timeit
    def df_names_json(self):
        """
        A JSON object containing the names of the dataframes
        """
        return json.dumps({'df_names': self.curr_step['df_names']})

    @property
    @timeit
    def sheet_json(self):
        """
        sheet_json contains a serialized representation of the data
        frames that is then fed into the ag-grid in the front-end. 

        NOTE: we only display the _first_ 2,000 rows of the dataframe
        for speed reasons. This results in way less data getting 
        passed around
        """
        return dfs_to_json(self.curr_step['dfs'])
    
    @property
    @timeit
    def df_shape_json(self):
        """
        Returns the df shape (rows, columns) of each dataframe in the 
        current step!
        """
        return json.dumps([
            {'rows': df.shape[0], 'cols': df.shape[1]}
            for df in self.curr_step['dfs']
        ])

    @property
    @timeit
    def column_spreadsheet_code_json(self):
        """
        column_spreadsheet_code_json is a list of all the spreadsheet
        formulas that users have used, for each sheet they have. 
        """
        return json.dumps(self.curr_step['column_spreadsheet_code'])

    @property
    @timeit
    def code_json(self):
        """
        This code json string is sent to the front-end and is what
        ends up getting displayed in the codeblock. 
        """
        return json.dumps(transpile(self))

    @property
    @timeit
    def column_filters_json(self):
        """
        This column_filters list is used by the front end to display
        the filtered icons in the UI
        """
        return json.dumps(self.curr_step['column_filters'])
    
    @property
    @timeit
    def column_type_json(self):
        """
        Returns a list of JSON objects that hold the type of each
        data in each column.
        """
        return json.dumps(self.curr_step['column_type'])

    @timeit
    def handle_edit_event(self, edit_event):
        """
        Updates the widget state with a new step that was created
        by the edit_event. Each edit_event creates at most one step. 

        If there is an error in the creation of the new step, this
        function will delete the new, invalid step.
        """

        pre_update_step_num = len(self.steps) 

        for new_step in STEPS:
            if edit_event['type'] == new_step['event_type']:
                # Get the params for this event
                params = {key: value for key, value in edit_event.items() if key in new_step['params']}
                # Actually execute this event
                try:
                    new_step['execute'](self, **params)
                except:
                    # If there is _any_ error while executing this event, we:
                    # 1. Delete the new step, if it has been created at all.
                    # 2. Bubble the error up
                    if len(self.steps) > pre_update_step_num:
                        self.steps = self.steps[:pre_update_step_num]

                    # https://stackoverflow.com/questions/6593922/letting-an-exception-to-bubble-up
                    raise

                # Return if execution is sucessful
                return

        # If we didn't find anything, then we error!
        raise Exception(f'{edit_event} is not an edit event!')

    @timeit
    def handle_update_event(self, update_event):
        """
        Handles any event that isn't caused by an edit, but instead
        other types of new data coming from the frontend (e.g. the df names 
        or some existing steps).
        """
        if update_event['type'] == 'df_names_update':
            df_names = update_event['df_names']
            self.curr_step['df_names'] = deepcopy(df_names)
        elif update_event['type'] == 'save_analysis':
            analysis_name = update_event['analysis_name']
            write_analysis(self, analysis_name)
        elif update_event['type'] == 'use_existing_analysis_update':
            step_summaries = update_event['steps']
            self._rerun_analysis(step_summaries)
        else:
            for update in UPDATES:
                if update_event['type'] == update['event_type']:
                    # Get the params for this event
                    params = {key: value for key, value in update_event.items() if key in update['params']}
                    # Actually execute this event
                    update['execute'](self, **params)
                    # And then return
                    return

            raise Exception(f'{update_event} is not an update event!')

    def _rerun_analysis(self, step_summaries):
        """
        This function reapplies all the steps summarized in the passed step summaries, 
        which come from a saved analysis. 

        If any of the step summaries fails, this function tries to roll back to before
        it applied any of the stems
        """   
        # We make a shallow copy of the steps, as none of the objects
        # will be changed by the step summaries we apply   
        old_steps = copy(self.steps)  
        
        try:
            for step_id, step_summary in step_summaries.items():
                step_type = step_summary['step_type']

                found = False
                for new_step in STEPS:
                    if step_type == new_step['step_type']:
                        found = True
                        # Get the params for this event
                        params = {key: value for key, value in step_summary.items() if key in new_step['params']}
                        # Actually execute this event
                        new_step['execute'](self, **params)

                if not found:
                    raise Exception('Trying to recreate invalid step:', step_summary)

        except Exception as e:
            print(e)
            # We remove all applied steps if there was an error
            self.steps = old_steps

            # And report a generic error to the user
            raise make_execution_error()


    def _create_and_checkout_new_step(self, step_type):
        """
        Creates a new step with new_step_id and step_type that starts
        with the ending state of the previous step
        """
        new_step_id = self.curr_step_id + 1

        # the new step is a copy of the previous step, where we only take the data we need
        # (which is the formula content only)
        new_step = dict()

        new_step['step_type'] = step_type
        new_step['column_metatype'] = deepcopy(self.steps[new_step_id - 1]['column_metatype'])
        new_step['column_type'] = deepcopy(self.steps[new_step_id - 1]['column_type'])
        new_step['column_spreadsheet_code'] = deepcopy(self.steps[new_step_id - 1]['column_spreadsheet_code'])
        new_step['added_columns'] = [[] for df in self.steps[new_step_id - 1]['dfs']]
        new_step['column_python_code'] = deepcopy(self.steps[new_step_id - 1]['column_python_code'])
        new_step['column_evaluation_graph'] = deepcopy(self.steps[new_step_id - 1]['column_evaluation_graph'])
        new_step['column_filters'] = deepcopy(self.steps[new_step_id - 1]['column_filters'])
        new_step['dfs'] = deepcopy(self.steps[new_step_id - 1]['dfs'])
        new_step['df_names'] = deepcopy(self.steps[new_step_id - 1]['df_names'])

        # add the new step to list of steps
        self.steps.append(new_step)

    def _delete_curr_step(self):
        """
        Deletes the current step and rolls back a step!
        """
        self.steps.pop()

    def add_df_to_curr_step(self, new_df, df_name=None):
        """
        Helper function for adding a new dataframe to the current step!
        """
        # update dfs by appending new df
        self.curr_step['dfs'].append(new_df)
        # Also update the dataframe name
        if df_name is None:
            self.curr_step['df_names'].append(f'df{len(self.curr_step["df_names"]) + 1}')
        else:
            self.curr_step['df_names'].append(df_name)

        # Update all the variables that depend on column_headers
        column_headers = new_df.keys()
        self.curr_step['column_metatype'].append({column_header: 'value' for column_header in column_headers})
        self.curr_step['column_type'].append({column_header: get_column_filter_type(new_df[column_header]) for column_header in column_headers})
        self.curr_step['column_spreadsheet_code'].append({column_header: '' for column_header in column_headers})
        self.curr_step['column_python_code'].append({column_header: '' for column_header in column_headers})
        self.curr_step['column_evaluation_graph'].append({column_header: set() for column_header in column_headers})
        self.curr_step['column_filters'].append({column_header: {'operator':'And', 'filters': []} for column_header in column_headers})

    def does_sheet_index_exist(self, index):
        return not (index < 0 or index >= self.num_sheets)
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A delete_column step, which allows you to delete a column
from a dataframe.
"""

from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_column_delete_error
)

DELETE_COLUMN_EVENT = 'delete_column_edit'
DELETE_COLUMN_STEP_TYPE = 'delete_column'

DELETE_COLUMN_PARAMS = [
    'sheet_index', # the sheet to delete the column from
    'column_header', # the column to delete
]

def execute_delete_column_step(
        wsc,
        sheet_index,
        column_header
    ):
    """
    Creates a new step that deletes a the column column_heder 
    from the sheet at sheet_index.
    """

    # if the sheet doesn't exist, throw an error
    if not wsc.does_sheet_index_exist(sheet_index):
        raise make_no_sheet_error(sheet_index)

    # Error if the column does not exist
    if column_header not in wsc.curr_step['column_metatype'][sheet_index]:
        raise make_no_column_error([column_header])
    
    # Error if there are any columns that currently rely on this column
    if len(wsc.curr_step['column_evaluation_graph'][sheet_index][column_header]) > 0:
        raise make_invalid_column_delete_error(
            column_header,
            list(wsc.curr_step['column_evaluation_graph'][sheet_index][column_header])
        )

    # Make a new step for the delete
    wsc._create_and_checkout_new_step(DELETE_COLUMN_STEP_TYPE)
    # Save the parameters
    wsc.curr_step['sheet_index'] = sheet_index
    wsc.curr_step['column_header'] = column_header

    # Actually drop the column
    df = wsc.curr_step['dfs'][sheet_index]
    df.drop(column_header, axis=1, inplace=True)

    # And then update all the state variables removing this column from the state
    del wsc.curr_step['column_metatype'][sheet_index][column_header]
    del wsc.curr_step['column_type'][sheet_index][column_header]
    del wsc.curr_step['column_spreadsheet_code'][sheet_index][column_header]
    del wsc.curr_step['column_python_code'][sheet_index][column_header]
    del wsc.curr_step['column_evaluation_graph'][sheet_index][column_header]
    # We also have to delete the places in the graph where this node is 
    for dependents in wsc.curr_step['column_evaluation_graph'][sheet_index].values():
        if column_header in dependents:
            dependents.remove(column_header)


def transpile_delete_column_step(
        widget_state_container,
        step,
        sheet_index,
        column_header
    ):
    df_name = step['df_names'][sheet_index]
    return [f'{df_name}.drop(\'{column_header}\', axis=1, inplace=True)']


DELETE_COLUMN_STEP = {
    'step_version': 1,
    'event_type': DELETE_COLUMN_EVENT,
    'step_type': DELETE_COLUMN_STEP_TYPE,
    'params': DELETE_COLUMN_PARAMS,
    'execute': execute_delete_column_step,
    'transpile': transpile_delete_column_step
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A set_column_formula step, which allows you to set the formula
of a given column in the sheet (and then recalculates this column)
and it's dependents.
"""
from mitosheet.topological_sort import creates_circularity, topological_sort_columns
from mitosheet.sheet_functions import FUNCTIONS
from mitosheet.parser import parse_formula
from mitosheet.utils import get_column_filter_type

from mitosheet.errors import (
    make_circular_reference_error, 
    make_no_column_error, 
    make_unsupported_function_error, 
    make_wrong_column_metatype_error
)

SET_COLUMN_FORMULA_EVENT = 'set_column_formula_edit'
SET_COLUMN_FORMULA_STEP_TYPE = 'set_column_formula'

SET_COLUMN_FORMULA_PARAMS = [
    'sheet_index', # int
    'column_header', # the new column to create
    'old_formula', # the previous formula, or None (if there is None)
    'new_formula', # the new formula for the column
]


def _update_column_formula_in_curr_step(
        wsc,
        sheet_index,
        column_header,
        old_formula,
        new_formula
    ):
    """
    A  helper function for updating the formula of a column. It assumes
    that the passed information is all _correct_ and will not:
    1. Introduce a circular reference error.
    2. Add an invalid formula.

    It DOES NOT "reexecute" the dataframes, just updates the state variables.
    """

    new_python_code, _, new_dependencies = parse_formula(
        new_formula, 
        column_header
    )

    _, _, old_dependencies = parse_formula(
        old_formula, 
        column_header
    )

    wsc.curr_step['column_spreadsheet_code'][sheet_index][column_header] = new_formula
    wsc.curr_step['column_python_code'][sheet_index][column_header] = new_python_code

    # Update the column dependency graph
    for old_dependency in old_dependencies:
        wsc.curr_step['column_evaluation_graph'][sheet_index][old_dependency].remove(column_header)
    for new_dependency in new_dependencies:
        wsc.curr_step['column_evaluation_graph'][sheet_index][new_dependency].add(column_header)


def execute_set_column_formula_step(
        wsc,
        sheet_index,
        column_header,
        old_formula,
        new_formula
    ):
    """
    Sets the column with column_header to have the new_formula, and 
    updates the dataframe as a result.

    Errors if:
    - The given column_header is not a column. 
    - The new_formula introduces a circular reference.
    - The new_formula causes an execution error in any way. 

    In the case of an error, this function rolls back all variables
    variables to their state at the start of this function.
    """
    if column_header not in wsc.curr_step['column_metatype'][sheet_index]:
        raise make_no_column_error([column_header])

    # First, we check the column_metatype, and make sure it's a formula
    if wsc.curr_step['column_metatype'][sheet_index][column_header] != 'formula':
        raise make_wrong_column_metatype_error(column_header)

    # If nothings changed, there's no work to do
    if (old_formula == new_formula):
        return

    # Then we try and parse the formula
    new_python_code, new_functions, new_dependencies = parse_formula(
        new_formula, 
        column_header
    )

    # We check that the formula doesn't reference any columns that don't exist
    missing_columns = new_dependencies.difference(wsc.curr_step['column_metatype'][sheet_index].keys())
    if any(missing_columns):
        raise make_no_column_error(missing_columns)

    # The formula can only reference known formulas
    missing_functions = new_functions.difference(set(FUNCTIONS.keys()))
    if any(missing_functions):
        raise make_unsupported_function_error(missing_functions)

    # Then, we get the list of old column dependencies and new dependencies
    # so that we can update the graph
    old_python_code, old_functions, old_dependencies = parse_formula(old_formula, column_header)

    # Before changing any variables, we make sure this edit didn't
    # introduct any circularity
    circularity = creates_circularity(
        wsc.curr_step['column_evaluation_graph'][sheet_index], 
        column_header,
        old_dependencies,
        new_dependencies
    )
    if circularity:
        raise make_circular_reference_error()

    # We check out a new step, and save the params there
    wsc._create_and_checkout_new_step(SET_COLUMN_FORMULA_STEP_TYPE)
    wsc.curr_step['sheet_index'] = sheet_index
    wsc.curr_step['column_header'] = column_header
    wsc.curr_step['old_formula'] = old_formula
    wsc.curr_step['new_formula'] = new_formula

    # Update the column formula, and then execute the new formula graph
    _update_column_formula_in_curr_step(wsc, sheet_index, column_header, old_formula, new_formula)
    _execute(wsc, wsc.curr_step['dfs'][sheet_index], sheet_index)

    # Finially, update the type of the filters of this column, for all the filters
    new_type = get_column_filter_type(wsc.curr_step['dfs'][sheet_index][column_header])
    wsc.curr_step['column_type'][sheet_index][column_header] = new_type
    wsc.curr_step['column_filters'][sheet_index][column_header]['filters'] = [
        {'type': new_type, 'condition': filter_['condition'], 'value': filter_['value']} 
        for filter_ in wsc.curr_step['column_filters'][sheet_index][column_header]['filters']
    ]

def _execute(wsc, df, sheet_index):
    """
    Executes the given state variables for  
    """

    topological_sort = topological_sort_columns(wsc.curr_step['column_evaluation_graph'][sheet_index])

    for column_header in topological_sort:
        # Exec the code, where the df is the original dataframe
        # See explination here: https://www.tutorialspoint.com/exec-in-python
        exec(
            wsc.curr_step['column_python_code'][sheet_index][column_header],
            {'df': df}, 
            FUNCTIONS
        )


def transpile_set_column_formula_step(
        widget_state_container,
        step,
        sheet_index,
        column_header,
        old_formula,
        new_formula
    ):
    """
    Transpiles an set_column_formula step to python code!
    """
    code = []

    # We only look at the sheet that was changed, and sort it's columns
    topological_sort = topological_sort_columns(step['column_evaluation_graph'][sheet_index])

    # For all columns downstream of the changed column
    topological_sort = topological_sort[topological_sort.index(column_header):]

    # We compile all of their formulas
    for column in topological_sort:
        column_formula_changes = step['column_python_code'][sheet_index][column]
        if column_formula_changes != '':
            # We replace the data frame in the code with it's parameter name!
            column_formula_changes = column_formula_changes.strip().replace('df', f'{step["df_names"][sheet_index]}')
            code.append(column_formula_changes)

    return code


SET_COLUMN_FORMULA_STEP = {
    'step_version': 1,
    'event_type': SET_COLUMN_FORMULA_EVENT,
    'step_type': SET_COLUMN_FORMULA_STEP_TYPE,
    'params': SET_COLUMN_FORMULA_PARAMS,
    'execute': execute_set_column_formula_step,
    'transpile': transpile_set_column_formula_step
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains all functions that don't particularly fit in any other category

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
import pandas as pd
import numpy as np
import math

from mitosheet.sheet_function_types import as_types
from mitosheet.sheet_functions.sheet_function_utils import fill_series_list_to_max

@as_types('series', 'number')
def OFFSET(series: pd.Series, offset) -> pd.Series:
    """
    {
        "function": "OFFSET",
        "description": "Shifts the given series by the given offset. Use a negative offset to reference a previous row, and a offset number to reference a later row.",
        "examples": [
            "OFFSET(Nums, 10)",
            "OFFSET(call_time, -1)"
        ],
        "syntax": "OFFSET(series, offset)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to shift up or down."
            },
            {
                "element": "offset",
                "description": "An integer amount to shift. Use a negative number to reference a previous row, and a positive number to reference a later row."
            }
        ]
    }
    """
    length = len(series)
    # If the shift is too big, we just return an empty series
    if abs(offset) >= length:
        return pd.Series([np.NaN] * length)
    
    if offset < 0:
        # Otherwise, append the NaNs to the front, and chop at the correct length
        return pd.concat([pd.Series([np.NaN] * (offset * -1)), series], ignore_index=True).head(length)
    else:
        remaining_series = series.tail(length - offset)
        return pd.concat([remaining_series, pd.Series([np.NaN] * offset)], ignore_index=True).head(length)


@as_types('series', 'series')
def FILLNAN(series: pd.Series, replacement) -> pd.Series:
    """
    {
        "function": "FILLNAN",
        "description": "Replaces the NaN values in the series with the replacement value.",
        "examples": [
            "FILLNAN(A, 10)",
            "FILLNAN(A, 'replacement')"
        ],
        "syntax": "FILLNAN(series, replacement)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to replace the NaN values in."
            },
            {
                "element": "replacement",
                "description": "A string, number, or date to replace the NaNs with."
            }
        ]
    }
    """
    nan_indexes = series.isna()
    for idx in range(len(nan_indexes)):
        if nan_indexes[idx]:
            series[idx] = replacement[0]

    return series

# TODO: we should see if we can list these automatically!
MISC_FUNCTIONS = {
    'OFFSET': OFFSET,
    'FILLNAN': FILLNAN
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Information about the frontend package of the widgets.
"""

module_name = 'mitosheet'
module_version = '^0.1.67'



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A filter steps, which allows you to filter a column based on some
conditions and some values. 

NOTE: there are some workarounds in this step with old filter steps being deleted
and remade (like the one right) - simply because we don't have a way of editing
old steps otherwise!
"""
from copy import deepcopy
import functools
from numbers import Number
import pandas as pd

from mitosheet.errors import (
    EditError,
    make_execution_error,
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_filter_error
)

FILTER_COLUMN_EVENT = 'filter_column_edit'
FILTER_COLUMN_STEP_TYPE = 'filter_column'

FILTER_COLUMN_PARAMS = [
    'sheet_index', # int
    'column_header', # column to filter from
    'operator', # 'Or' or 'And'
    'filters', # the trippples of ({'type': <type>, 'condition': <condition>, 'value': <value>})
]


# CONSTANTS USED IN THE FILTER STEP ITSELF

STRING_TYPE = 'string'
NUMBER_TYPE = 'number'
DATETIME_TYPE = 'datetime'

SHARED_FILTER_CONDITIONS = [
    'empty',
    'not_empty'
]

STRING_FILTER_CONDITIONS = [
    'contains',
    'string_does_not_contain',
    'string_exactly'
]

NUMBER_FILTER_CONDITIONS = [
    'number_exactly',
    'greater',
    'greater_than_or_equal',
    'less',
    'less_than_or_equal'
]

DATETIME_FILTER_CONDITIONS = [
    'datetime_exactly',
    'datetime_greater',
    'datetime_greater_than_or_equal',
    'datetime_less',
    'datetime_less_than_or_equal',
]


def execute_filter_column(
        wsc,
        sheet_index,
        column_header,
        operator,
        filters
    ):
    """
    Filters an existing sheet with the given filters, which 
    each contain a condition and an optional value.
    """
    # if the sheet doesn't exist, throw an error
    if not wsc.does_sheet_index_exist(sheet_index):
        raise make_no_sheet_error(sheet_index)

    # We check that the filtered column exists 
    missing_column = set([column_header]).difference(wsc.curr_step['column_metatype'][sheet_index].keys())
    if len(missing_column) > 0: 
        raise make_no_column_error(missing_column)

    # if the condition is none, don't create a new step and delete the current filter if it exists
    if len(filters) == 0:
        # if there is a filter already applied to the column, remove it
        _reset_filter(wsc, sheet_index, column_header, ignore_last_step=False)
        return

    # If no errors we create a new step for this filter
    wsc._create_and_checkout_new_step(FILTER_COLUMN_STEP_TYPE)
    # update the step to save the variables needed to reconstruct the filter
    wsc.curr_step['sheet_index'] = sheet_index
    wsc.curr_step['column_header'] = column_header
    wsc.curr_step['filters'] = filters
    wsc.curr_step['operator'] = operator

    # Then we update the dataframe, first by executing on a fake dataframe
    try:
        # TODO: Remove this speculative execution, when it's safe to do so
        # (e.g. when we have proper step editing). 
        # For now, because filter is wacky, we leave it in here... 

        # make a copy of our data frame to test operate on 
        df_copy = deepcopy(wsc.curr_step['dfs'][sheet_index])

        # We execute on the copy first to see if there will be errors
        _execute_filter(
            df_copy, 
            column_header,
            operator,
            filters
        )

    except EditError as e:
        # If an edit error occurs, we delete the filter step
        wsc._delete_curr_step()
        # And we propagate this error upwards
        raise e
    except Exception as e:
        print(e)
        # If any other error occurs, we delete the filter step
        wsc._delete_curr_step()
        # We raise a generic execution error in this case!
        raise make_execution_error()

    # if there is a filter already applied to the column, remove it
    # NOTE: we ignore the last step as _the step we are currently executing_ 
    # would get deleted otherwise, which we do not want. 
    _reset_filter(wsc, sheet_index, column_header, ignore_last_step=True)

    # if there was no error in execution on the copy, execute on real dataframes
    wsc.dfs[sheet_index] = _execute_filter(
        wsc.dfs[sheet_index], 
        column_header,
        operator,
        filters
    )

    # keep track of which columns are filtered
    wsc.curr_step['column_filters'][sheet_index][column_header]['operator'] = operator
    wsc.curr_step['column_filters'][sheet_index][column_header]['filters'] = filters

def _reset_filter(wsc, sheet_index, column_header, ignore_last_step=False):
    """
    To reset the filter, we:

    1. Delete the filter step.
    2. Save the analysis.
    3. Delete _all_ the steps (ik, crazy).
    4. Replay the analysis.

    NOTE: we do this as a workaround because we want users to be able to edit
    filters, but there is currently no way to go back to a step and edit it. 

    We should _heavily_ re-evaluate this screwy-ness when we allow editing of steps, 
    and rolling back to steps.
    """
    # NOTE: we import here to avoid circular imports. This is an unfortunate
    # side effect of this entire workaround :( )
    from mitosheet.save_utils import read_analysis, write_analysis

    deleted_step_id = None

    # find the step to delete
    for step_id, step in enumerate(wsc.steps):
        if ignore_last_step and step_id == wsc.curr_step_id:
            continue

        if step['step_type'] == FILTER_COLUMN_STEP_TYPE:
            # If this is the step that added a filter to the column, mark it as the one to delete
            if step['sheet_index'] == sheet_index and step['column_header'] == column_header:
                deleted_step_id = step_id

    # if there are no step to delete, return
    if deleted_step_id == None:
        return
    
    # Delete the filter step
    wsc.steps.pop(deleted_step_id)

    # save the current analysis, now without the filter step
    write_analysis(wsc)

    # Delete _all the steps_ except the first
    wsc.steps = [wsc.steps[0]]

    # Rerun the analysis
    saved_analysis = read_analysis(wsc.analysis_name)
    wsc._rerun_analysis(saved_analysis['steps'])

def _execute_filter(
        df, 
        column_header,
        operator,
        filters
    ):
    """
    Executes a filter on the given column, filtering by removing any rows who
    don't meet the condition.
    """

    applied_filters = []

    for filter_ in filters:
        type_ = filter_['type']
        condition = filter_['condition']
        value = filter_['value']

        # First, we case on the shared conditions, to get them out of the way
        if condition in SHARED_FILTER_CONDITIONS:
            if condition == 'empty':
                applied_filters.append(df[column_header].isna())
            elif condition == 'not_empty':
                applied_filters.append(df[column_header].notnull())
            continue

        if type_ == STRING_TYPE:
            if condition not in STRING_FILTER_CONDITIONS:
                raise Exception(f'Invalid condition passed to string filter {condition}')

            # Check that the value is the valid
            if not isinstance(value, str):
                raise make_invalid_filter_error(value, STRING_TYPE)

            if condition == 'contains':
                applied_filters.append(df[column_header].str.contains(value, na=False))
            if condition == 'string_does_not_contain':
                applied_filters.append(~df[column_header].str.contains(value, na=False))
            elif condition == 'string_exactly':
                applied_filters.append(df[column_header] == value)

        elif type_ == NUMBER_TYPE:
            if condition not in NUMBER_FILTER_CONDITIONS:
                raise Exception(f'Invalid condition passed to number filter {condition}')
            
            # Check that the value is the valid
            if not isinstance(value, Number):
                raise make_invalid_filter_error(value, NUMBER_TYPE)

            if condition == 'number_exactly':
                applied_filters.append(df[column_header] == value)
            elif condition == 'greater':
                applied_filters.append(df[column_header] > value)
            elif condition == 'greater_than_or_equal':
                applied_filters.append(df[column_header] >= value)
            elif condition == 'less':
                applied_filters.append(df[column_header] < value)
            elif condition == 'less_than_or_equal':
                applied_filters.append(df[column_header] <= value)

        elif type_ == DATETIME_TYPE:
            if condition not in DATETIME_FILTER_CONDITIONS:
                raise Exception(f'Invalid condition passed to datetime filter {condition}')

            # Check that we were given something that can be understood as a date
            try:
                timestamp = pd.to_datetime(value)
            except:
                raise make_invalid_filter_error(value, DATETIME_TYPE)

            if condition == 'datetime_exactly':
                applied_filters.append(df[column_header] == timestamp)
            elif condition == 'datetime_greater':
                applied_filters.append(df[column_header] > timestamp)
            elif condition == 'datetime_greater_than_or_equal':
                applied_filters.append(df[column_header] >= timestamp)
            elif condition == 'datetime_less':
                applied_filters.append(df[column_header] < timestamp)
            elif condition == 'datetime_less_than_or_equal':
                applied_filters.append(df[column_header] <= timestamp)

        else:
            raise Exception(f'Invalid type passed in filter {type_}')

    def filter_reducer(filter_one, filter_two):
        # helper for combining filters based on the operations
        if operator == 'Or':
            return (filter_one) | (filter_two)
        elif operator == 'And':
            return (filter_one) & (filter_two)
        else:
            raise Exception(f'Operator {operator} is unsupported')

    # Combine all the filters into a single filter
    final_filter = functools.reduce(filter_reducer, applied_filters)
    
    return df[final_filter].reset_index(drop=True)


def transpile_filter_column(
        widget_state_container,
        step,
        sheet_index,
        column_header,
        operator,
        filters
    ):
    """
    Transpiles a filter step to Python code. 
    """
    df_name = step['df_names'][sheet_index]

    # build the filter code
    partial_filter_code = []
    for filter_ in filters:
        condition = filter_['condition']
        value = filter_['value']

        FILTER_FORMAT_STRING_DICT = {
            # SHARED CONDITIONS
            'empty': '{df_name}.{column_header}.isna()',
            'not_empty': '{df_name}.{column_header}.notnull()',

            # NUMBERS
            'number_exactly': '{df_name}[\'{column_header}\'] == {value}',
            'greater': '{df_name}[\'{column_header}\'] > {value}',
            'greater_than_or_equal': '{df_name}[\'{column_header}\'] >= {value}',
            'less': '{df_name}[\'{column_header}\'] < {value}',
            'less_than_or_equal': '{df_name}[\'{column_header}\'] <= {value}',
            
            # STRINGS
            'contains': '{df_name}[\'{column_header}\'].str.contains(\'{value}\', na=False)',
            'string_does_not_contain': '~{df_name}[\'{column_header}\'].str.contains(\'{value}\', na=False)',
            'string_exactly': '{df_name}[\'{column_header}\'] == \'{value}\'',

            # DATES
            'datetime_exactly': '{df_name}[\'{column_header}\'] == pd.to_datetime(\'{value}\')',
            'datetime_greater': '{df_name}[\'{column_header}\'] > pd.to_datetime(\'{value}\')',
            'datetime_greater_than_or_equal': '{df_name}[\'{column_header}\'] >= pd.to_datetime(\'{value}\')',
            'datetime_less': '{df_name}[\'{column_header}\'] < pd.to_datetime(\'{value}\')',
            'datetime_less_than_or_equal': '{df_name}[\'{column_header}\'] <= pd.to_datetime(\'{value}\')',            
        }

        if condition in FILTER_FORMAT_STRING_DICT:
            partial_filter_code.append(
                FILTER_FORMAT_STRING_DICT[condition].format(
                    df_name=df_name,
                    column_header=column_header,
                    value=value
                )
            )
        else:
            continue

    if len(partial_filter_code) == 0:
        return []
    elif len(partial_filter_code) == 1:
        return [
            f'{df_name} = {df_name}[{partial_filter_code[0]}]',
            f'{df_name} = {df_name}.reset_index(drop=True)'
        ]
    else:
        # If there are multiple conditions, we combine them together, with the
        # given operator in the middle
        OPERATOR_SIGNS = {
            'Or': '|',
            'And': '&'
        }
        filter_string = f' {OPERATOR_SIGNS[operator]} '.join([f'({pfc})' for pfc in partial_filter_code])
        return [
            f'{df_name} = {df_name}[{filter_string}]',
            f'{df_name} = {df_name}.reset_index(drop=True)'
        ]


FILTER_STEP = {
    'step_version': 1,
    'event_type': FILTER_COLUMN_EVENT,
    'step_type': FILTER_COLUMN_STEP_TYPE,
    'params': FILTER_COLUMN_PARAMS,
    'execute': execute_filter_column,
    'transpile': transpile_filter_column
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A group step, which allows you to group data
from an existing dataframe along some keys, and then
aggregate other columns with specific functions.
"""
from pandas.core.base import DataError
import pandas as pd

from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_aggregation_error,
    make_invalid_aggregation_error,
)

GROUP_EVENT = 'group_edit'
GROUP_STEP_TYPE = 'group'

GROUP_PARAMS = [
    'sheet_index', # int
    'group_rows', # list of column_headers, could be empty
    'group_columns', # list of column_headers, could be empty
    'values', # a dict from column_header -> aggregation function
]

def execute_group_step(
        wsc,
        sheet_index,
        group_rows,
        group_columns,
        values
    ):
    """
    The function responsible for updating the widget state container
    with a new group step.

    If it fails part of the way through, deletes the new group step entirely.
    """

    # if the sheets don't exist, throw an error
    if not wsc.does_sheet_index_exist(sheet_index):
        raise make_no_sheet_error(sheet_index)

    # We check that the group by doesn't use any columns that don't exist
    columns_used = set(group_rows).union(set(group_columns))
    missing_group_keys = columns_used.difference(wsc.curr_step['column_metatype'][sheet_index].keys())
    if len(missing_group_keys) > 0:
        raise make_no_column_error(missing_group_keys)

    missing_value_keys = set(values.keys()).difference(wsc.curr_step['column_metatype'][sheet_index].keys())
    if len(missing_value_keys) > 0:
        raise make_no_column_error(missing_value_keys)

    # Create a new step
    wsc._create_and_checkout_new_step(GROUP_STEP_TYPE)
    # Save the params of the current step
    wsc.curr_step['sheet_index'] = sheet_index
    wsc.curr_step['group_rows'] = group_rows
    wsc.curr_step['group_columns'] = group_columns
    wsc.curr_step['values'] = values

    # Actually execute the grouping
    try:
        new_df = _execute_group(
            wsc.curr_step['dfs'][sheet_index], 
            group_rows,
            group_columns,
            values
        )
    except DataError as e:
        # A data-error occurs when you try to aggregate on a column where the function
        # cannot be applied (e.g. 'mean' on a column of strings)
        print(e)
        # Generate an error informing the user
        raise make_invalid_aggregation_error()

    # Add it to the dataframe
    wsc.add_df_to_curr_step(new_df)

def _execute_group(df, group_rows, group_columns, values):
    """
    Helper function for executing the groupby on a specific dataframe
    and then aggregating the values with the passed values mapping
    """

    # First, we handle a special case where the only row is also the value
    # the user wants to count. In this case, we
    # This is something that C.M. wants + needs :) 
    if len(group_rows) == 1 and len(group_columns) == 0 and len(values) == 1 and values == {group_rows[0]: 'count'}:
        groupby_obj = df.groupby([group_rows[0]], as_index=False)
        return groupby_obj.size()


    values_keys = list(values.keys())

    # Built the args, leaving out any unused values
    args = {}

    if len(group_rows) > 0:
        args['index'] = group_rows

    if len(group_columns) > 0:
        args['columns'] = group_columns

    if len(values) > 0:
        args['values'] = values_keys
        args['aggfunc'] = values

    pivot_table = df.pivot_table(**args) # type: pd.DataFrame

    # Flatten the column headers
    pivot_table.columns = [
        '_'.join([str(c) for c in col]).strip() if isinstance(col, tuple) else col
        for col in pivot_table.columns.values
    ]

    # flatten the column headers & reset the indexes
    pivot_table = pivot_table.rename_axis(None, axis=1).reset_index()

    return pivot_table

# Helpful constants for code formatting
TAB = '    '
NEWLINE_TAB = f'\n{TAB}'


def build_args_code(
        group_rows,
        group_columns,
        values
    ):
    """
    Helper function for building an arg string, while leaving
    out empty arguments. 
    """
    values_keys = list(values.keys())

    args = []
    if len(group_rows) > 0:
        args.append(f'index={group_rows},')

    if len(group_columns) > 0:
        args.append(f'columns={group_columns},')

    if len(values) > 0:
        args.append(f'values={values_keys},')
        args.append(f'aggfunc={values}')

    return NEWLINE_TAB.join(args)


def transpile_group_step(
        widget_state_container,
        step,
        sheet_index,
        group_rows,
        group_columns,
        values
    ):
    """
    Transpiles a group step to python code!
    """
    new_df_name = f'df{len(step["dfs"])}'


    # First, we handle a special case where the only row is also the value the user wants to count. 
    # In this case, we simply count it! This handles a common case users bump into
    if len(group_rows) == 1 and len(group_columns) == 0 and len(values) == 1 and values == {group_rows[0]: 'count'}:
        return [
            f'groupby_obj = {step["df_names"][sheet_index]}.groupby([\'{group_rows[0]}\'], as_index=False)',
            f'{new_df_name} = groupby_obj.size()'
        ]

    transpiled_code = []
    
    # Pivot 
    pivot_comment = '# Pivot the data'
    pivot_table_args = build_args_code(group_rows, group_columns, values)
    pivot_table_call = f'pivot_table = {step["df_names"][sheet_index]}.pivot_table({NEWLINE_TAB}{pivot_table_args}\n)'
    transpiled_code.append(f'{pivot_comment}\n{pivot_table_call}')

    # Flatten column headers. 
    # NOTE: this only needs to happen if there is more than one group_column, as the columns
    # will be flat if there are no columns grouped
    if len(group_columns) > 0:
        flatten_comment = '# Flatten the column headers'
        flatten_code = f'pivot_table.columns = [{NEWLINE_TAB}\'_\'.join([str(c) for c in col]).strip() if isinstance(col, tuple) else col{NEWLINE_TAB}for col in pivot_table.columns.values\n]'
        transpiled_code.append(f'{flatten_comment}\n{flatten_code}')

    # Finially, reset the column name, and the indexes!
    reset_index_comment = f'# Reset the column name and the indexes'
    reset_index_code = f'{new_df_name} = pivot_table.rename_axis(None, axis=1).reset_index()'
    transpiled_code.append(f'{reset_index_comment}\n{reset_index_code}')

    return transpiled_code


GROUP_STEP = {
    'step_version': 1,
    'event_type': GROUP_EVENT,
    'step_type': GROUP_STEP_TYPE,
    'params': GROUP_PARAMS,
    'execute': execute_group_step,
    'transpile': transpile_group_step
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Generates a JSON list of function documentation objects from doc strings
of sheet functions in mitosheet/sheet_functions.py. See docs/README for 
a description of this object and more information.
"""
import json

def main():
    """
    Reads in all the sheet functions' documentation objects, 
    and writes them to src/data/function_documentation.json.
    """

    from mitosheet.sheet_functions import FUNCTIONS

    function_doc_objects = []
    invalid = []
    for function_name, function in FUNCTIONS.items():
        try:
            func_doc_obj = json.loads(function.__doc__)
            function_doc_objects.append(func_doc_obj)
            print(
                f'Build documentation for {function_name}!'
            )
        except:
            print(function.__doc__)
            # If we can't read in the function documentation, report that
            invalid.append(function_name)

    if len(invalid) > 0:
        invalid_str = ', '.join(invalid)
        print(f'\nUnable to generate documentation for {len(invalid)} functions: {invalid_str}. Make sure they are valid JSON objects as specificed in ./docs/README.md.')

    # Sort them alphabetically
    function_doc_objects.sort(key=lambda doc: doc['function'])

    with open('./src/data/function_documentation.tsx', 'w+') as f:
        # We write the types for these objects as well!
        f.write("""
export interface FunctionDocumentationObject {
    function: string;
    description: string;
    examples?: (string)[] | null;
    syntax: string;
    syntax_elements?: (SyntaxElementsEntity)[] | null;
}

export interface SyntaxElementsEntity {
    element: string;
    description: string;
}

export const functionDocumentationObjects: FunctionDocumentationObject[] = """)

        f.write(json.dumps(function_doc_objects))

    print('\nWrote documentation to ./src/data/function_documentation.tsx!')

    

if __name__ == '__main__':
    main()


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
TODO: Add module docstring
"""
from mitosheet.api import handle_api_event
import sys
import analytics
import pandas as pd
from typing import List
import traceback

from ipywidgets import DOMWidget
import traitlets as t

from ._frontend import module_name, module_version
from .errors import EditError, get_recent_traceback_as_list, log_recent_error
from .save_utils import (
    read_analysis, 
    write_analysis,
    saved_analysis_names_json
)
from mitosheet.widget_state_container import WidgetStateContainer
from mitosheet.profiling import timeit

from mitosheet.mito_analytics import analytics, static_user_id

class MitoWidget(DOMWidget):
    """TODO: Add docstring here
    """
    _model_name = t.Unicode('ExampleModel').tag(sync=True)
    _model_module = t.Unicode(module_name).tag(sync=True)
    _model_module_version = t.Unicode(module_version).tag(sync=True)
    _view_name = t.Unicode('ExampleView').tag(sync=True)
    _view_module = t.Unicode(module_name).tag(sync=True)
    _view_module_version = t.Unicode(module_version).tag(sync=True)

    value = t.Unicode('Hello World').tag(sync=True)
    analysis_name = t.Unicode('').tag(sync=True)
    sheet_json = t.Unicode('').tag(sync=True)
    code_json = t.Unicode('').tag(sync=True)
    df_names_json = t.Unicode('').tag(sync=True)
    df_shape_json = t.Unicode('').tag(sync=True)
    saved_analysis_names_json = t.Unicode('').tag(sync=True)
    user_id = t.Unicode(static_user_id).tag(sync=True)
    column_spreadsheet_code_json = t.Unicode('').tag(sync=True)
    column_filters_json = t.Unicode('').tag(sync=True)
    column_type_json = t.Unicode('').tag(sync=True)


    def __init__(self, *args, **kwargs):
        """
        Takes a list of dataframes, passed through *args. 

        NOTE: assumes that the passed arguments are all dataframes, 
        and also have all valid headers. These conditions are checked in the
        mitosheet.sheet function. 
        """
        # Call the DOMWidget constructor to set up the widget properly
        super(MitoWidget, self).__init__()

        # Set up the state container to hold private widget state
        self.widget_state_container = WidgetStateContainer(args)

        # Set up starting shared state variables
        self.update_shared_state_variables()

        # Set up message handler
        self.on_msg(self.receive_message)

    def update_shared_state_variables(self):
        """
        Helper function for updating all the variables that are shared
        between the backend and the frontend through trailets.
        """
        self.sheet_json = self.widget_state_container.sheet_json
        self.column_spreadsheet_code_json = self.widget_state_container.column_spreadsheet_code_json
        self.code_json = self.widget_state_container.code_json
        self.df_names_json = self.widget_state_container.df_names_json
        self.df_shape_json = self.widget_state_container.df_shape_json
        self.analysis_name = self.widget_state_container.analysis_name
        self.saved_analysis_names_json = saved_analysis_names_json()
        self.column_filters_json = self.widget_state_container.column_filters_json
        self.column_type_json = self.widget_state_container.column_type_json

    @timeit 
    def send(self, msg):
        """
        We overload the DOMWidget's send function, so that 
        we log all outgoing messages
        """
        # Send the message though the DOMWidget's send function
        super().send(msg)

    @timeit
    def saturate(self, event):
        """
        Saturation is when the server fills in information that
        is missing from the event client-side. This is for consistency
        and because the client may not have easy access to the info
        all the time.
        """
        curr_step = self.widget_state_container.curr_step

        if event['event'] == 'edit_event':
            if event['type'] == 'set_column_formula_edit':
                sheet_index = event['sheet_index']
                column_header = event['column_header']
                event['old_formula'] = curr_step['column_spreadsheet_code'][sheet_index][column_header]
        if event['event'] == 'update_event':
            if event['type'] == 'use_existing_analysis_update':
                # If we're getting an event telling us to update, we read in the steps from the file
                analysis_name = event['analysis_name']
                analysis = read_analysis(analysis_name)
                for key, value in analysis.items():
                    event[key] = value

                # When replaying an analysis with import events, you can also send over
                # new params to the import events to replace them. We replace them in the steps here
                if 'import_summaries' in event:
                    for step_id, params in event['import_summaries'].items():
                        for key, value in params.items():
                            event['steps'][step_id][key] = value
                
        if event['event'] == 'api_call':
            handle_api_event(self.send, event)

    @timeit
    def handle_edit_event(self, event):
        """
        Handles an edit_event. Per the spec, an edit_event
        updates both the sheet and the codeblock, and as such
        the sheet is re-evaluated and the code for the codeblock
        is re-transpiled.

        Useful for any event that changes the state of both the sheet
        and the codeblock!
        """
        # First, we log the edit that we just got
        analytics.track(self.user_id, event['type'], event)

        # First, we send this new edit to the evaluator
        self.widget_state_container.handle_edit_event(event)

        # We update the state variables 
        self.update_shared_state_variables()

        # Also, write the analysis to a file!
        write_analysis(self.widget_state_container)

        # Tell the front-end to render the new sheet and new code
        self.send({"event": "update_sheet"})
        self.send({"event": "update_code"})


    @timeit
    def handle_update_event(self, event):
        """
        This event is not the user editing the sheet, but rather information
        that has been collected from the frontend (after render) that is being
        passed back.

        For example:
        - Names of the dataframes
        - Name of an existing analysis
        """
        # First, we log the update event that we just got
        analytics.track(self.user_id, event['type'], event)

        self.widget_state_container.handle_update_event(event)

        # Update all state variables and send out an update
        self.update_shared_state_variables()

        # Also, write the analysis to a file!
        write_analysis(self.widget_state_container)

        # Tell the front-end to render the new sheet and new code
        self.send({"event": "update_sheet"})
        self.send({"event": "update_code"})

    def receive_message(self, widget, content, buffers=None):
        """
        Handles all incoming messages from the JS widget. There are two main
        types of events:

        1. edit_event: any event that updates the state of the sheet and the
        code block at once. Leads to reevaluation, and a re-transpile.

        2. update_event: any event that isn't caused by an edit, but instead
        other types of new data coming from the frontend (e.g. the df names 
        or some existing steps).

        3. A log_event is just an event that should get logged on the backend.
        """
        # First, we saturate the event
        event = content
        self.saturate(event)

        try:
            if event['event'] == 'edit_event':
                self.handle_edit_event(event)
            if event['event'] == 'update_event':
                self.handle_update_event(event)
            if event['event'] == 'log_event':
                analytics.track(
                    self.user_id, 
                    event['type'], 
                    event
                )
        except EditError as e:
            print(e)
            # If we hit an error during editing, log that it has occured
            analytics.track(
                self.user_id, 
                f'{e.type_}_log_event', 
                {
                    'header': e.header, 
                    'to_fix': e.to_fix,
                    'traceback': get_recent_traceback_as_list()
                }
            )
            # Report it to the user, and then return
            self.send({
                'event': 'edit_error',
                'type': e.type_,
                'header': e.header,
                'to_fix': e.to_fix
            })
        except:
            # We log the error
            log_recent_error()
            # Report it to the user, and then return
            self.send({
                'event': 'edit_error',
                'type': 'execution_error',
                'header': 'Execution Error',
                'to_fix': 'Sorry, there was an error during executing this code.'
            })

@timeit
def sheet(*args: List[pd.DataFrame]) -> MitoWidget:
    """
    Returns a Mito Widget defined on the passed data frames.

    Note: errors if any of the arguments are not data frames
    """
    args = list(args)

    for df in args:
        if not isinstance(df, pd.DataFrame):
            # Log the error
            analytics.track(
                static_user_id, 
                'sheet_error_log_event', 
                {'error': f'Invalid argument passed to sheet: {df}. Please pass all dataframes.'}
            )

            raise ValueError(f'Invalid argument passed to sheet: {df}. Please pass all dataframes.')

    try:
        widget = MitoWidget(*args) 
    except:
        # We log the error
        log_recent_error('sheet_error_log_event')
        # And then bubble it to the user
        raise

    return widget
